# -*- coding: utf-8 -*-

"""Module to enhance LaneletNetwork class
so it can be used for conversion from the opendrive format."""

from typing import Tuple, List
from queue import Queue
import numpy as np

from commonroad.scenario.lanelet import LaneletNetwork
from .conversion_lanelet import ConversionLanelet

__author__ = "Benjamin Orthen"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


def convert_to_new_lanelet_id(old_lanelet_id: str, ids_assigned: dict) -> int:
    """Convert the old lanelet ids (format 501.1.-1.-1) to newer,
    simpler ones (100, 101 etc.).

    Do this by consecutively assigning
    numbers, starting at 100, to the old_lanelet_id strings. Save the
    assignments in the dict which is passed to the function as ids_assigned.

    Args:
      old_lanelet_id: Old id with format "501.1.-1.-1".
      ids_assigned: Dict with all previous assignments

    Returns:
      The new lanelet id.

    """

    starting_lanelet_id = 100

    if old_lanelet_id in ids_assigned.keys():
        new_lanelet_id = ids_assigned[old_lanelet_id]
    else:
        try:
            new_lanelet_id = max(ids_assigned.values()) + 1
        except ValueError:
            new_lanelet_id = starting_lanelet_id
        ids_assigned[old_lanelet_id] = new_lanelet_id

    return new_lanelet_id


class ConversionLaneletNetwork(LaneletNetwork):
    """Add functions to LaneletNetwork which
    further enable it to modify its Lanelets."""

    def remove_lanelet(self, lanelet_id: str, remove_references: bool = False):
        """Remove a lanelets with the specific lanelet_id
        from the _lanelets dict.

        Args:
          lanelet_id: id of lanelet to be removed.
          remove_references: Also remove references which point to to be removed lanelet.

        Returns:
          None

        """
        del self._lanelets[lanelet_id]
        if remove_references:
            for lanelet in self.lanelets:
                lanelet.predecessor[:] = [
                    pred for pred in lanelet.predecessor if pred != lanelet_id
                ]

                lanelet.successor[:] = [
                    succ for succ in lanelet.successor if succ != lanelet_id
                ]
                if lanelet.adj_right == lanelet_id:
                    lanelet.adj_right = None
                if lanelet.adj_left == lanelet_id:
                    lanelet.adj_left = None

    def find_lanelet_by_id(self, lanelet_id) -> ConversionLanelet:
        """Find a lanelet for a given lanelet_id.
        Disable natural number check of parent class.

        Args:
          lanelet_id: The id of the lanelet to find

        Returns:
          The lanelet object if the id exists and None otherwise

        """
        return self._lanelets.get(lanelet_id)

    def convert_all_lanelet_ids(self):
        """Convert lanelet ids to numbers which comply with the Commonroad specification.

        These numbers have to be positive integers.
        """

        new_lanelet_ids_assigned = {}

        for lanelet in self.lanelets:
            lanelet.description = lanelet.lanelet_id
            self.remove_lanelet(lanelet.lanelet_id)
            lanelet.lanelet_id = convert_to_new_lanelet_id(
                lanelet.lanelet_id, new_lanelet_ids_assigned
            )
            lanelet.predecessor = [
                convert_to_new_lanelet_id(x, new_lanelet_ids_assigned)
                for x in lanelet.predecessor
            ]
            lanelet.successor = [
                convert_to_new_lanelet_id(x, new_lanelet_ids_assigned)
                for x in lanelet.successor
            ]
            if lanelet.adj_left is not None:
                lanelet.adj_left = convert_to_new_lanelet_id(
                    lanelet.adj_left, new_lanelet_ids_assigned
                )
            if lanelet.adj_right is not None:
                lanelet.adj_right = convert_to_new_lanelet_id(
                    lanelet.adj_right, new_lanelet_ids_assigned
                )
            self.add_lanelet(lanelet)

    def prune_network(self):
        """Remove references in predecessor, successor etc. to
        non existing lanelets.

        Args:

        Returns:

        """
        self.delete_zero_width_parametric_lanes()

        lanelet_ids = [x.lanelet_id for x in self.lanelets]

        for lanelet in self.lanelets:
            lanelet.predecessor[:] = [
                pred for pred in lanelet.predecessor if pred in lanelet_ids
            ]
            lanelet.successor[:] = [
                succ for succ in lanelet.successor if succ in lanelet_ids
            ]
            if lanelet.adj_left not in lanelet_ids:
                lanelet.adj_left = None
            if lanelet.adj_right not in lanelet_ids:
                lanelet.adj_right = None

    def delete_zero_width_parametric_lanes(self):
        """Remove all ParametricLaneGroup which have zero width at every point from
        this network.
        """
        for lanelet in self.lanelets:
            if lanelet.has_zero_width_everywhere():
                if lanelet.adj_right:
                    adj_right = self.find_lanelet_by_id(lanelet.adj_right)
                    if adj_right:
                        if adj_right.adj_left == lanelet.lanelet_id:
                            adj_right.adj_left = lanelet.adj_left
                            adj_right.adj_left_same_direction = (
                                lanelet.adj_left_same_direction
                            )
                        else:
                            adj_right.adj_right = lanelet.adj_left
                            adj_right.adj_right_same_direction = (
                                not lanelet.adj_left_same_direction
                            )

                if lanelet.adj_left:
                    adj_left = self.find_lanelet_by_id(lanelet.adj_left)
                    if adj_left:
                        if adj_left.adj_right == lanelet.lanelet_id:
                            adj_left.adj_right = lanelet.adj_right
                            adj_left.adj_right_same_direction = (
                                lanelet.adj_right_same_direction
                            )
                        else:
                            adj_left.adj_left = lanelet.adj_right
                            adj_left.adj_left_same_direction = (
                                not lanelet.adj_right_same_direction
                            )

                self.remove_lanelet(lanelet.lanelet_id, remove_references=True)

    def update_lanelet_id_references(self, old_id: str, new_id: str):
        """Update all references to the old lanelet_id with the new_lanelet_id.

        Args:
          old_id: Old lanelet_id which has changed.
          new_id: New lanelet_id the old_id has changed into.

        """

        for lanelet in self.lanelets:
            lanelet.predecessor[:] = [
                new_id if pred == old_id else pred for pred in lanelet.predecessor
            ]

            lanelet.successor[:] = [
                new_id if succ == old_id else succ for succ in lanelet.successor
            ]

            if lanelet.adj_right == old_id:
                lanelet.adj_right = new_id

            if lanelet.adj_left == old_id:
                lanelet.adj_left = new_id

    def concatenate_possible_lanelets(self):
        """Iterate trough lanelets in network and concatenate possible lanelets together.

        Check for each lanelet if it can be concatenated with its successor and
        if its neighbors can be concatenated as well. If yes, do the concatenation.

        """
        concatenate_lanelets = []
        for lanelet in self.lanelets:

            possible_concat_lanes = None
            if lanelet.adj_right is None:
                possible_concat_lanes = self.check_concatenation_potential(
                    lanelet, "left"
                )
            elif lanelet.adj_left is None:
                possible_concat_lanes = self.check_concatenation_potential(
                    lanelet, "right"
                )
            if possible_concat_lanes:
                concatenate_lanelets.append(possible_concat_lanes)

        # this dict saves the lanelet_ids which point to another lanelet_id
        # because they were concatenated together and the resulting lanelet
        # can only have one lanelet_id.
        # key in dict has been renamed to value
        replacement_ids = dict()

        for possible_concat_lanes in concatenate_lanelets:
            # prevent chains of more than one lanelet being renamed
            replacement_ids = {
                k: replacement_ids.get(v, v) for k, v in replacement_ids.items()
            }

            possible_concat_lanes = [
                (
                    replacement_ids.get(pair[0], pair[0]),
                    replacement_ids.get(pair[1], pair[1]),
                )
                for pair in possible_concat_lanes
            ]
            replacement_ids.update(
                self._concatenate_lanelet_pairs_group(possible_concat_lanes)
            )

    def _concatenate_lanelet_pairs_group(self, lanelet_pairs: list) -> dict:
        """Concatenate a group of lanelet_pairs, with setting correctly the new lanelet_ids
        at neighbors.

        Args:
          lanelet_pairs: List with tuples of lanelet_ids which should be concatenated.

        Returns:
          Dict with information which lanelet_id was converted to a new one.

        """

        new_lanelet_ids = dict()
        for pair in lanelet_pairs:
            if pair[0] == pair[1]:
                continue
            lanelet_1 = self.find_lanelet_by_id(pair[0])
            lanelet_2 = self.find_lanelet_by_id(pair[1])
            lanelet_1.concatenate(lanelet_2)

            self.remove_lanelet(pair[1])

            # each reference to lanelet_2 should point to lanelet_id
            # of new_lanelet instead
            self.update_lanelet_id_references(
                lanelet_2.lanelet_id, lanelet_1.lanelet_id
            )
            # update dict to show which lanelet_id changed to which
            new_lanelet_ids[pair[1]] = pair[0]

        return new_lanelet_ids

    def join_and_split_possible_lanes(self):
        """Move lanelet boundaries for lanelet splits or joins.

        This method provides the functionality to modify the lane boundaries if
        a lane merges into another lane or splits from another lane.
        """
        # Condition for lane merge:
        # left and right vertices at end or beginning are the same
        js_targets = []
        for lanelet in self.lanelets:

            lanelet_split, lanelet_join = False, False
            if not lanelet.predecessor and np.allclose(
                    lanelet.left_vertices[0], lanelet.right_vertices[0]
            ):
                lanelet_split = True

            if not lanelet.successor and np.allclose(
                    lanelet.left_vertices[-1], lanelet.right_vertices[-1]
            ):
                lanelet_join = True

            if lanelet_join or lanelet_split:
                js_targets.append(
                    _JoinSplitTarget(self, lanelet, lanelet_split, lanelet_join)
                )

        for js_target in js_targets:
            js_target.determine_apt_js_pairs()
            js_target.move_borders()
            js_target.add_adjacent_predecessor_or_successor()

    def predecessor_is_neighbor_of_neighbors_predecessor(
            self, lanelet: "ConversionLanelet"
    ) -> bool:
        """Checks if neighbors of predecessor are the successor of the adjacent neighbors
        of the lanelet.

        Args:
          lanelet: Lanelet to check neighbor requirement for.

        Returns:
          True if this neighbor requirement is fulfilled.

        """
        if not self.has_unique_pred_succ_relation(-1, lanelet):
            return False

        predecessor = self.find_lanelet_by_id(lanelet.predecessor[0])
        return self.successor_is_neighbor_of_neighbors_successor(predecessor)

    def add_successors_to_lanelet(
            self, lanelet: ConversionLanelet, successor_ids: List[str]
    ):
        """Add a successor to a lanelet, but add the lanelet also to the predecessor
        of the succesor.

        Args:
          lanelet: Lanelet to add successor to.
          successor_ids: Id of successor to add to lanelet.
        """
        for successor_id in successor_ids:
            lanelet.successor.append(successor_id)
            successor = self.find_lanelet_by_id(successor_id)
            successor.predecessor.append(lanelet.lanelet_id)

    def add_predecessors_to_lanelet(
            self, lanelet: ConversionLanelet, predecessor_ids: List[str]
    ):
        """Add a successor to a lanelet, but add the lanelet also to the predecessor
        of the succesor.

        Args:
          lanelet: Lanelet to add successor to.
          predecessor_id: Id of successor to add to lanelet.
        """
        for predecessor_id in predecessor_ids:
            lanelet.predecessor.append(predecessor_id)
            predecessor = self.find_lanelet_by_id(predecessor_id)
            predecessor.successor.append(lanelet.lanelet_id)

    def set_adjacent_left(
            self, lanelet: ConversionLanelet, adj_left_id: str, same_direction: bool = True
    ):
        """Set the adj_left of a lanelet to a new value.

        Update also the lanelet which is the new adjacent left to
        have new adjacent right.

        Args:
          lanelet: Lanelet which adjacent left should be updated.
          adj_left_id: New value for update.
          same_direction: New adjacent lanelet has same direction as lanelet.
        Returns:
          True if operation successful, else false.
        """
        new_adj = self.find_lanelet_by_id(adj_left_id)
        if not new_adj:
            return False
        lanelet.adj_left = adj_left_id
        lanelet.adj_left_same_direction = same_direction
        if same_direction:
            new_adj.adj_right = lanelet.lanelet_id
            new_adj.adj_right_same_direction = True
        else:
            new_adj.adj_left = lanelet.lanelet_id
            new_adj.adj_left_same_direction = False
        return True

    def set_adjacent_right(
            self, lanelet: ConversionLanelet, adj_right_id: str, same_direction: bool = True
    ):
        """Set the adj_right of a lanelet to a new value.

        Update also the lanelet which is the new adjacent right to
        have new adjacent left.

        Args:
          lanelet: Lanelet which adjacent right should be updated.
          adj_right_id: New value for update.
          same_direction: New adjacent lanelet has same direction as lanelet.
        Returns:
          True if operation successful, else false.
        """
        new_adj = self.find_lanelet_by_id(adj_right_id)
        if not new_adj:
            return False
        lanelet.adj_right = adj_right_id
        lanelet.adj_right_same_direction = same_direction
        if same_direction:
            new_adj.adj_left = lanelet.lanelet_id
            new_adj.adj_left_same_direction = True
        else:
            new_adj.adj_right = lanelet.lanelet_id
            new_adj.adj_right_same_direction = False
        return True

    def check_concatenation_potential(
            self, lanelet: ConversionLanelet, adjacent_direction: str
    ) -> list:
        """Check if lanelet could be concatenated with its successor.

        Args:
          lanelet: Lanelet to check concatenation potential with its successor
          adjacent_direction: "Left" or "Right", determinating which lanelet

        Returns:
          A list of pairs of lanelets which can be concatenated. None if it is not possible.

        """
        mergeable_lanelets = []
        neighbor_ok = self.successor_is_neighbor_of_neighbors_successor(lanelet)
        if not neighbor_ok:
            return None

        if adjacent_direction == "left":
            mergeable_lanelets.append((lanelet.lanelet_id, lanelet.successor[0]))
            if lanelet.adj_left is None:
                return mergeable_lanelets
            if lanelet.adj_left_same_direction:
                next_neighbor = self.find_lanelet_by_id(lanelet.adj_left)
                new_direction = "left"
            else:
                next_neighbor = self.find_lanelet_by_id(
                    self.find_lanelet_by_id(lanelet.adj_left).predecessor[0]
                )
                new_direction = "right"

        else:
            mergeable_lanelets.append((lanelet.lanelet_id, lanelet.successor[0]))
            if lanelet.adj_right is None:
                return mergeable_lanelets
            if lanelet.adj_right_same_direction:
                next_neighbor = self.find_lanelet_by_id(lanelet.adj_right)
                new_direction = "right"
            else:
                next_neighbor = self.find_lanelet_by_id(
                    self.find_lanelet_by_id(lanelet.adj_right).predecessor[0]
                )
                new_direction = "left"

        recursive_merge = self.check_concatenation_potential(
            next_neighbor, new_direction
        )
        if recursive_merge is None:
            return None

        mergeable_lanelets.extend(recursive_merge)
        return mergeable_lanelets

    def successor_is_neighbor_of_neighbors_successor(
            self, lanelet: ConversionLanelet
    ) -> bool:
        """Checks if neighbors of successor are the successor of the adjacent neighbors
        of the lanelet.

        Args:
          lanelet: Lanelet to check specified relation for.

        Returns:
          True if this neighbor requirement is fulfilled.

        """
        if not self.has_unique_pred_succ_relation(1, lanelet):
            return False

        return self.adj_left_consistent_nb(lanelet) and self.adj_right_consistent_nb(
            lanelet
        )

    def has_unique_pred_succ_relation(
            self, direction: int, lanelet: ConversionLanelet
    ) -> bool:
        """Checks if lanelet has only one successor/predecessor and the
        successor/predecessor has only one predecessor/successor, s.t.
        it is a one-to-one relation.

        Args:
          direction: 1 if the successor should be checked.
        -1 (or all other values) for the predecessor.
          lanelet_network: Network to search for lanelets by ids.
          direction: int:
          lanelet_network: "LaneletNetwork":

        Returns:
          True if the relation is unique, False otherwise.

        """
        if direction == 1:
            neighbors = lanelet.successor
        else:
            neighbors = lanelet.predecessor

        # check if neighbor is only one
        if neighbors is None or len(neighbors) != 1:
            return False

        # get lanelet object with id
        neighbor = self.find_lanelet_by_id(neighbors[0])

        # get the neighbor of the neighbor
        nb_neighbor = neighbor.predecessor if direction == 1 else neighbor.successor

        # check if nb_neighbor has one neighbor in proper direction
        if nb_neighbor is None or len(nb_neighbor) != 1:
            return False

        # return True if these ids are the same (they should be!)
        return lanelet.lanelet_id == nb_neighbor[0]

    def adj_right_consistent_nb(self, lanelet: ConversionLanelet) -> bool:
        """Checks if right neighbor of successor is the successor
        of the right adjacent neighbor of the lanelet.

        Args:
          lanelet: Lanelet to check specified relation for.

        Returns:
          True if this neighbor requirement is fulfilled.

        """
        successor = self.find_lanelet_by_id(lanelet.successor[0])
        adj_right = self.find_lanelet_by_id(lanelet.adj_right)
        if adj_right:
            if lanelet.adj_right_same_direction:
                if not self.has_unique_pred_succ_relation(1, adj_right):
                    return False
                if adj_right.successor[0] != successor.adj_right:
                    return False
            else:
                if not lanelet.has_unique_pred_succ_relation(-1, adj_right):
                    return False
                if adj_right.predecessor[0] != successor.adj_right:
                    return False
        else:
            return successor.adj_right is None
        return True

    def adj_left_consistent_nb(self, lanelet: ConversionLanelet) -> bool:
        """Checks if left neighbor of successor is the successor of the
        left adjacent neighbor of the lanelet.

        Args:
          lanelet: Lanelet to check specified relation for.

        Returns:
          True if this neighbor requirement is fulfilled.

        """
        successor = self.find_lanelet_by_id(lanelet.successor[0])
        adj_left = self.find_lanelet_by_id(lanelet.adj_left)
        if adj_left:
            if lanelet.adj_left_same_direction:
                if not self.has_unique_pred_succ_relation(1, adj_left):
                    return False
                if adj_left.successor[0] != successor.adj_left:
                    return False
            else:
                if not self.has_unique_pred_succ_relation(-1, adj_left):
                    return False
                if adj_left.predecessor[0] != successor.adj_left:
                    return False
        else:
            return successor.adj_left is None
        return True


class _JoinSplitTarget:
    """Class to integrate joining/splitting of lanelet borders.

    Provides methods to determine the lanelets with which the
    join and/or split can be performed. Additionally a method to
    change the borders of the determined lanelets.

    Attributes:
      main_lanelet (ConversionLanelet): Lanelet where split starts or join ends.
      lanelet_network (ConversionLaneletNetwork): LaneletNetwork where join/split occurs.
      _mode (int): Number denoting if join (0), split (1), or join and split (2) occurs.
      change_width (float): Width at start of split or end of join.
        Is list with two elements, [split_width, join_width] if _mode == 2
      linking_side (str): Side on which the split/join happens (either "left" or "right")
      _js_pairs (list): List of :class:`._JoinSplitPair` elements.
      _single_lanelet_operation (bool): Indicates whether only one lanelet and
        its adjacent lanelet can be used for the join/split.
    """

    def __init__(
            self,
            lanelet_network: ConversionLaneletNetwork,
            main_lanelet: ConversionLanelet,
            split: bool,
            join: bool,
    ):
        self.main_lanelet = main_lanelet
        self.lanelet_network = lanelet_network
        if split and join:
            self._mode = 2
        elif split:
            self._mode = 1
        else:
            self._mode = 0
        self.change_width = None
        self.linking_side = None
        self._js_pairs = []
        self._single_lanelet_operation = False

    @property
    def split(self):
        """Lanelet splits at start.

        Returns:
          True if lanelet splits from other lanelet at start.
        """
        return self._mode >= 1

    @property
    def join(self):
        """Lanelet joins at end.

        Returns:
          True if lanelet joins to other lanelet at end.
        """
        return self._mode != 1

    @property
    def split_and_join(self) -> bool:
        """Lanelet splits at start and joins at end.

        Returns:
          True if it has a join and a split.
        """
        return self.split and self.join

    def use_only_single_lanelet(self) -> bool:
        """Only single lanelet can be used for join/split.

        Returns:
          True if only one can be used.
        """
        return self._single_lanelet_operation and self.split_and_join

    def _find_lanelet_by_id(self, lanelet_id: str) -> ConversionLanelet:
        """Run :func:`.ConversionLaneletNetwork.find_lanelet_by_id` of self.lanelet_network.

        Returns:
          Lanelet matching the lanelet_id.
        """
        return self.lanelet_network.find_lanelet_by_id(lanelet_id)

    def complete_js_interval_length(self) -> float:
        """Calculate length of interval where join/split changes the border.

        Returns:
          Length of interval.
        """
        length = 0
        for js_pair in self._js_pairs:
            length += js_pair.change_interval[1] - js_pair.change_interval[0]

        return length

    def adjacent_width(self, is_split: bool) -> float:
        """Get width of adjacent lanelet at start of split or end of join.

        Returns:
          Width of adjacent lanelet at start or end.
        """
        if is_split:
            return self._js_pairs[0].adjacent_lanelet.calc_width_at_start()
        return self._js_pairs[0].adjacent_lanelet.calc_width_at_end()

    def add_adjacent_predecessor_or_successor(self):
        """Add the predecessor or successor of the adjacent lanelet to the main lanelet.

        This reflects that after the split, the main lanelet is also a successor
        of the predecessor of its adjacent lanelet.
        For a join, the main lanelet is a predecessor of the successor of its
        adjacent lanelet.
        """

        if not self._js_pairs:
            return
        lanelet = self._js_pairs[0].lanelet
        adjacent_lanelet = self._js_pairs[0].adjacent_lanelet
        if self.split:
            self.lanelet_network.add_predecessors_to_lanelet(
                lanelet, adjacent_lanelet.predecessor
            )

        if self.join:
            self.lanelet_network.add_successors_to_lanelet(
                lanelet, adjacent_lanelet.successor
            )

    def move_borders(self):
        """Move borders of lanelets to reflect the split/join.

        All lanelet pairs in self._js_pairs are used for the border movement.
        """

        if not self._js_pairs:
            return

        if self.split_and_join:
            self._move_borders_if_split_and_join()
        else:
            self._move_borders_if_split_or_join()

    def _move_borders_if_split_or_join(self):
        """Move borders of lanelets if it is not split and join.

        Interpolate width interval for each js_pair.
        Then move the borders of its lanelet.
        """
        length = self.complete_js_interval_length()
        adj_width = self.adjacent_width(is_split=self.split)
        if self.join:
            js_pairs = list(reversed(self._js_pairs))
            # norm running position so that running_pos + pos_start
            # is at zero at first js_pair
            running_pos = -1 * self._js_pairs[0].change_interval[0]
            width_start = self.change_width
            width_end = adj_width
        else:
            js_pairs = self._js_pairs
            running_pos = 0
            width_start = adj_width
            width_end = self.change_width
        for js_pair in js_pairs:
            [pos_start, pos_end] = js_pair.change_interval
            distance = np.interp(
                [pos_start + running_pos, pos_end + running_pos],
                [0, length],
                [width_start, width_end],
            )
            js_pair.move_border(width=distance, linking_side=self.linking_side)
            running_pos += pos_end - pos_start

    def _move_borders_if_split_and_join(self):
        """Move borders of lanelets if it is split and join.

        Calculate the new vertices twice:
        1. Only for the split (first pair in self._js_pairs)
        2. Only for the join (second pair in self._js_pairs)
        Then talk the first half of the vertices of the split and
        the seconds half of the vertices of the join and merge them.
        """
        lanelet = self._js_pairs[0].lanelet

        start_width_split = self.adjacent_width(is_split=True)
        lanelet_split = self._js_pairs[0].move_border(
            width=[start_width_split, self.change_width[0]],
            linking_side=self.linking_side,
        )
        left_vertices = lanelet_split.left_vertices
        right_vertices = lanelet_split.right_vertices
        center_vertices = lanelet_split.center_vertices
        start_width_join = self.adjacent_width(is_split=False)
        self._js_pairs[1].move_border(
            width=[self.change_width[1], start_width_join],
            linking_side=self.linking_side,
        )

        # take first half of lanelet which does the split
        # take second half of lanelet which does the join
        half_length = int(left_vertices[:, 0].size / 2)
        lanelet.left_vertices = np.vstack(
            (left_vertices[:half_length, :], lanelet.left_vertices[half_length:, :])
        )
        lanelet.right_vertices = np.vstack(
            (right_vertices[:half_length, :], lanelet.right_vertices[half_length:, :])
        )
        lanelet.center_vertices = np.vstack(
            (center_vertices[:half_length, :], lanelet.center_vertices[half_length:, :])
        )

    def determine_apt_js_pairs(self):
        """Determine pairs of lanelet and adjacent lanelet for the join/split.

        Add lanelets as long as one of the break conditions is not matched.
        The determined pairs are saved in self._js_pairs.
        """
        # for first lanelet
        adjacent_lanelet = self._determine_main_adjacent_lanelet()
        if not adjacent_lanelet:
            return
        lanelet = self.main_lanelet
        while True:
            algo_has_finished = self._add_join_split_pair(lanelet, adjacent_lanelet)
            if algo_has_finished or self.use_only_single_lanelet():
                break
            if (
                    self.split
                    and self.lanelet_network.successor_is_neighbor_of_neighbors_successor(
                lanelet
            )
                    # and lanelet.successor[0] not in global_adjacent_lanelets
            ):
                lanelet = self._find_lanelet_by_id(lanelet.successor[0])
                adjacent_lanelet = self._find_lanelet_by_id(
                    adjacent_lanelet.successor[0]
                )
            elif self.join and (
                    self.lanelet_network.predecessor_is_neighbor_of_neighbors_predecessor(
                        lanelet
                    )
                    # and lanelet.predecessor[0] not in global_adjacent_lanelets
            ):
                lanelet = self._find_lanelet_by_id(lanelet.predecessor[0])
                adjacent_lanelet = self._find_lanelet_by_id(
                    adjacent_lanelet.predecessor[0]
                )

            else:
                break

    def _add_join_split_pair(
            self, lanelet: ConversionLanelet, adjacent_lanelet: ConversionLanelet
    ) -> bool:
        """Add a pair of lanelet and adjacent lanelet to self._js_pairs.

        Decide if it is advisable to add another pair to increase join/split area.

        Args:
          lanelet: Lanelet to be added.
          adjacent_lanelet: Lanelet adjacent to lanelet to be added.
        Returns:
          Indicator whether this was the last pair to be added. False means
           it is advisable to add another lanelet pair.
        """
        if self.split_and_join:
            # one for split at start of lanelet
            change_pos, change_width = lanelet.optimal_join_split_values(
                is_split=True,
                split_and_join=self.split_and_join,
                reference_width=adjacent_lanelet.calc_width_at_start(),
            )
            self._js_pairs.append(
                _JoinSplitPair(lanelet, adjacent_lanelet, [0, change_pos])
            )
            self.change_width = [change_width]
            # one for join at the end of the lanelet
            change_pos, change_width = lanelet.optimal_join_split_values(
                is_split=False,
                split_and_join=self.split_and_join,
                reference_width=adjacent_lanelet.calc_width_at_end(),
            )
            self._js_pairs.append(
                _JoinSplitPair(lanelet, adjacent_lanelet, [change_pos, lanelet.length])
            )
            self.change_width.append(change_width)
            return True

        adjacent_width = (
            adjacent_lanelet.calc_width_at_start()
            if self.split
            else adjacent_lanelet.calc_width_at_end()
        )
        change_pos, change_width = lanelet.optimal_join_split_values(
            is_split=self.split,
            split_and_join=self.split_and_join,
            reference_width=adjacent_width,
        )
        if self.change_width is not None and change_width < self.change_width:
            # algorithm to add lanelet should terminate
            return True
        self.change_width = change_width
        if self.split:
            self._js_pairs.append(
                _JoinSplitPair(lanelet, adjacent_lanelet, [0, change_pos])
            )
            if np.isclose(lanelet.length, change_pos):
                return False
        else:
            self._js_pairs.append(
                _JoinSplitPair(lanelet, adjacent_lanelet, [change_pos, lanelet.length])
            )
            if np.isclose(0, change_pos):
                return False
        return True

    def _determine_main_adjacent_lanelet(self) -> ConversionLanelet:
        """Determine which is the adjacent lanelet to the main lanelet.

        Returns:
          The corresponding adjacent lanelet.
        """
        lanelet = self.main_lanelet
        potential_adjacent_lanelets = Queue()
        checked_lanelets = 0
        if lanelet.adj_left is not None and lanelet.adj_left_same_direction:
            potential_adjacent_lanelets.put(
                {"lanelet_id": lanelet.adj_left, "linking_side": "right"}
            )
            checked_lanelets -= 1
        if lanelet.adj_right is not None and lanelet.adj_right_same_direction:
            potential_adjacent_lanelets.put(
                {"lanelet_id": lanelet.adj_right, "linking_side": "left"}
            )
            checked_lanelets -= 1

        while potential_adjacent_lanelets.qsize() > 0:
            adjacent_lanelet = self._check_next_adjacent_lanelet(
                potential_adjacent_lanelets
            )
            checked_lanelets += 1

            if checked_lanelets > 0:
                # adjacent lanelet is not next neighbor
                # successor of adjacent lanelet cant be used
                self._single_lanelet_operation = True
            if adjacent_lanelet is not None:
                # found appropriate adjacent lanelet
                return adjacent_lanelet

        return None

    def _check_next_adjacent_lanelet(
            self, potential_adjacent_lanelets: Queue
    ) -> ConversionLanelet:
        """Check next lanelet if it can act as adjacent lanelet to the main lanelet.

        If not, add its left and right neighbor, if they exists, to the potential_adjacent_lanelets Queue.

        Args:
          potential_adjacent_lanelets: Queue with dicts containing the pontential lanelets.
        Returns:
          Lanelet which fulfills the conditions if it exists, else None
        """
        adj_target = potential_adjacent_lanelets.get()
        adj_lanelet = self._find_lanelet_by_id(adj_target.get("lanelet_id"))
        linking_side = adj_target.get("linking_side")
        return_flag = 0

        if not adj_lanelet:
            return None
        if self.join:
            adj_width = adj_lanelet.calc_width_at_end()

        if self.split:
            adj_width = adj_lanelet.calc_width_at_start()
        if adj_width > 0:
            self.linking_side = adj_target.get("linking_side")
            return_flag = 1

        if return_flag:
            return adj_lanelet

        next_adj_neighbor = (
            adj_lanelet.adj_left if linking_side == "right" else adj_lanelet.adj_right
        )
        if next_adj_neighbor:
            potential_adjacent_lanelets.put(
                {"lanelet_id": next_adj_neighbor, "linking_side": linking_side}
            )
        return None


class _JoinSplitPair:
    "Pair of lanelet whose border is changed and its adjacent neighbor."

    def __init__(self, lanelet, adjacent_lanelet, change_interval):
        self.lanelet = lanelet
        self.adjacent_lanelet = adjacent_lanelet
        self.change_interval = change_interval

    def move_border(
            self, width: Tuple[float, float], linking_side: str
    ) -> ConversionLanelet:
        """Move border of self.lanelet.

        Args:
          width: Start and end value of new width of lanelet.
          linking_side: Side on which the split/join happens (either "left" or "right").
        Returns:
          Resulting lanelet after border movement.
        """
        self.lanelet.move_border(
            mirror_border=linking_side,
            mirror_interval=self.change_interval,
            distance=width,
            adjacent_lanelet=self.adjacent_lanelet,
        )
        return self.lanelet



# -*- coding: utf-8 -*-

"""Module describe parametric lane groups, which are groups combining consecutive
associated parametric lanes."""

from typing import Tuple
import math
import copy
import numpy as np

from ..conversion_lanelet import ConversionLanelet

__author__ = "Benjamin Orthen, Stefan Urban"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


class ParametricLaneGroup:
    """A group of parametric_lanes can be converted to a
    lanelet just like a single parametric_lane.
    """

    def __init__(
            self,
            id_=None,
            parametric_lanes=None,
            inner_neighbour=None,
            inner_neighbour_same_direction=True,
            outer_neighbour=None,
    ):

        self._geo_lengths = [np.array([0.0])]
        self.parametric_lanes = []
        self.id_ = id_
        self.inner_neighbour = inner_neighbour
        self.inner_neighbour_same_direction = inner_neighbour_same_direction
        self.outer_neighbour = outer_neighbour

        if parametric_lanes is not None:
            if isinstance(parametric_lanes, list):
                self.extend(parametric_lanes)
            else:
                self.append(parametric_lanes)

    def append(self, parametric_lane):
        """Append lane to start or end of interal list of ParametricLane objects.

        If the parametric_lane is reverse, it is inserted at the start.
        Args:
          parametric_lane: Lane to be inserted either at beginning or end of list.
        """
        if parametric_lane.reverse:
            self.parametric_lanes.insert(0, parametric_lane)
        else:
            self.parametric_lanes.append(parametric_lane)

        self._add_geo_length(parametric_lane.length, parametric_lane.reverse)

    def extend(self, plane_list: list):
        """Extend own ParametricLanes with new ones.

        Assumes ParametricLane objects in plane_list are already in order.

        Args:
          plane_list: List with ParametricLane objects.
        """
        for plane in plane_list:
            self.parametric_lanes.append(plane)
            self._add_geo_length(plane.length)

    def _add_geo_length(self, length: float, reverse: bool = False):
        """Add length of a ParametricLane to the array which keeps track
        at which position which ParametricLane is placed.

        This array is used for quickly accessing
        the proper ParametricLane for calculating a position.

        Args:
          length: Length of ParametricLane to be added.

        """
        if reverse:
            self._geo_lengths = np.insert(self._geo_lengths, 1, length)
            self._geo_lengths[2:] = [
                x + length for i, x in enumerate(self._geo_lengths) if i > 1
            ]
        else:
            self._geo_lengths = np.append(
                self._geo_lengths, length + self._geo_lengths[-1]
            )

    @property
    def type(self) -> str:
        """Get type of ParametricLaneGroup.


        Returns:
          Type of first ParametricLane in this Group.
        """
        return self.parametric_lanes[0].type_

    @property
    def length(self) -> float:
        """Length of all ParametricLanes which are collected in this ParametricLaneGroup.

        Returns:
          Accumulated length of ParametricLaneGroup.
        """

        return sum([x.length for x in self.parametric_lanes])

    def has_zero_width_everywhere(self) -> bool:
        """Checks if width is zero at every point of this ParametricLaneGroup.

        Returns:
          True if every ParametricLane has width_coefficients equal to only zero.
        """
        return all(
            [plane.has_zero_width_everywhere() for plane in self.parametric_lanes]
        )

    def to_lanelet(self, precision: float = 0.5, poses=None) -> ConversionLanelet:
        """Convert a ParametricLaneGroup to a Lanelet.

        Args:
          precision: Number which indicates at which space interval (in curve parameter ds)
            the coordinates of the boundaries should be calculated.
          mirror_border: Which lane to mirror, if performing merging or splitting of lanes.
          distance: Distance at start and end of lanelet, which mirroring lane should
            have from the other lane it mirrors.

        Returns:
          Created Lanelet.

        """
        left_vertices, right_vertices = np.array([]), np.array([])
        parametric_lane_poses = {}
        temp_parametric_lanes = sorted(self.parametric_lanes, key=lambda x: int(x.id_.split('.')[-1]))
        for parametric_lane in temp_parametric_lanes:
            if parametric_lane == temp_parametric_lanes[-1]:
                parametric_lane_poses[parametric_lane.id_] = poses
            else:
                for index in range(len(poses)):
                    if poses[index] > parametric_lane.length:
                        parametric_lane_poses[parametric_lane.id_] = copy.copy(poses[:index])
                        # poses 匹配width成功后移除匹配的元素, 减去 parametric_lane.length，否则后续可能长度不够用
                        poses = [i - parametric_lane.length for i in poses[index:]]
                        break
            # 未能在poses中匹配上自己所在的区间，取空值
            if not parametric_lane.id_ in parametric_lane_poses.keys():
                parametric_lane_poses[parametric_lane.id_] = []

        # self.parametric_lanes.sort(key=lambda x:x.id_) # 已经排序过了
        for parametric_lane in self.parametric_lanes:
            if not parametric_lane_poses[parametric_lane.id_]:
                # 通过 lane 点序列 获取 width 点序列， 特殊情况下，点序列可能为空 >> 原因是poses跨度较大，未能包含parametric_lane所在的区间
                continue
            local_left_vertices, local_right_vertices = parametric_lane.calc_vertices(
                precision=precision, poses=parametric_lane_poses[parametric_lane.id_]
            )

            if local_left_vertices is None:
                continue
            try:
                if np.isclose(left_vertices[-1], local_left_vertices[0]).all():
                    idx = 1
                else:
                    idx = 0
                left_vertices = np.vstack((left_vertices, local_left_vertices))
                right_vertices = np.vstack((right_vertices, local_right_vertices))
            except IndexError:
                left_vertices = local_left_vertices
                right_vertices = local_right_vertices

        center_vertices = np.array(
            [(l + r) / 2 for (l, r) in zip(left_vertices, right_vertices)]
        )

        lanelet = ConversionLanelet(
            copy.deepcopy(self), left_vertices, center_vertices, right_vertices, self.id_
        )

        # Adjacent lanes
        self._set_adjacent_lanes(lanelet)

        return lanelet

    def calc_border(self, border: str, s_pos: float, width_offset: float = 0.0):
        """Calc vertices point of inner or outer Border.

        Args:
          border: Which border to calculate (inner or outer).
          s_pos: Position of parameter ds where to calc the
        Cartesian coordinates
          width_offset: Offset to add to calculated width in reference
           to the reference border. (Default value = 0.0)

        Returns:
          Cartesian coordinates of point on inner border
            and tangential direction, too.

        """
        try:
            # get index of geometry which is at s_pos
            mask = self._geo_lengths > s_pos
            sub_idx = np.argmin(self._geo_lengths[mask] - s_pos)
            plane_idx = np.arange(self._geo_lengths.shape[0])[mask][sub_idx] - 1
        except ValueError:
            # s_pos is after last geometry because of rounding error
            if np.isclose(s_pos, self._geo_lengths[-1]):
                plane_idx = self._geo_lengths.size - 2
            else:
                raise Exception(
                    f"Tried to calculate a position outside of the borders of the reference path at s={s_pos}"
                    f", but path has only length of l={self._geo_lengths[-1]}"
                )

        return self.parametric_lanes[plane_idx].calc_border(
            border, s_pos - self._geo_lengths[plane_idx], width_offset
        )

    def to_lanelet_with_mirroring(
            self,
            mirror_border: str,
            distance: Tuple[float, float],
            mirror_interval: Tuple[float, float],
            adjacent_lanelet: ConversionLanelet,
            precision: float = 0.5,
    ):
        """Convert a ParametricLaneGroup to a Lanelet with mirroring one of the borders.

        Args:
          precision: Number which indicates at which space interval (in curve parameter ds)
            the coordinates of the boundaries should be calculated.
          mirror_border: Which lane to mirror, if performing merging or splitting of lanes.
          distance: Distance at start and end of lanelet, which mirroring lane should
            have from the other lane it mirrors.
          mirror_interval: Position at start and end of mirroring.

        Returns:
          Created Lanelet.

        """
        linear_distance_poly = np.polyfit(mirror_interval, distance, 1)
        distance_poly1d = np.poly1d(linear_distance_poly)
        global_distance = distance_poly1d([0, self.length])

        if self.parametric_lanes[0].reverse:
            global_distance[:] = [-x for x in global_distance]
        left_vertices, right_vertices = [], []

        last_width_difference = 0
        poses = self._calc_border_positions(precision)
        distance_slope = (global_distance[1] - global_distance[0]) / self.length
        for pos in poses:
            inner_pos = self.calc_border("inner", pos)[0]
            outer_pos = self.calc_border("outer", pos)[0]
            original_width = np.linalg.norm(inner_pos - outer_pos)

            # if not mirroring lane or outside of range
            if (
                    pos < mirror_interval[0] or pos > mirror_interval[1]
            ) and not np.isclose(pos, mirror_interval[1]):
                left_vertices.append(inner_pos)
                right_vertices.append(outer_pos)
                last_width_difference = 0

            else:
                # calculate positions of adjacent lanelet because new width of lanelet
                # cannot be more than width of adjacent lanelet and original width
                adj_inner_pos = adjacent_lanelet.calc_border("inner", pos)[0]
                adj_outer_pos = adjacent_lanelet.calc_border("outer", pos)[0]
                adjacent_width = np.linalg.norm(adj_inner_pos - adj_outer_pos)
                local_width_offset = distance_slope * pos + global_distance[0]

                if mirror_border == "left":
                    new_outer_pos = self.calc_border("inner", pos, local_width_offset)[
                        0
                    ]
                    modified_width = np.linalg.norm(new_outer_pos - inner_pos)

                    # change width s.t. it does not mirror inner border but instead
                    # outer border
                    local_width_offset = (
                            math.copysign(1, local_width_offset) * last_width_difference
                    )
                    if modified_width < original_width:
                        right_vertices.append(
                            self.calc_border("outer", pos, local_width_offset)[0]
                        )
                    elif modified_width > original_width + adjacent_width:
                        right_vertices.append(adj_outer_pos)
                    else:
                        right_vertices.append(new_outer_pos)
                        last_width_difference = abs(modified_width - original_width)

                    left_vertices.append(inner_pos)
                elif mirror_border == "right":
                    new_inner_pos = self.calc_border("outer", pos, local_width_offset)[
                        0
                    ]
                    modified_width = np.linalg.norm(new_inner_pos - outer_pos)

                    local_width_offset = (
                            math.copysign(1, local_width_offset) * last_width_difference
                    )
                    if modified_width < original_width:
                        left_vertices.append(
                            self.calc_border("inner", pos, local_width_offset)[0]
                        )
                    elif modified_width > original_width + adjacent_width:
                        left_vertices.append(adj_inner_pos)
                    else:
                        left_vertices.append(new_inner_pos)
                        last_width_difference = abs(modified_width - original_width)

                    right_vertices.append(outer_pos)

        left_vertices, right_vertices = (
            np.array(left_vertices),
            np.array(right_vertices),
        )
        # right_vertices = np.array(right_vertices)

        center_vertices = np.array(
            [(l + r) / 2 for (l, r) in zip(left_vertices, right_vertices)]
        )
        lanelet = ConversionLanelet(
            self, left_vertices, center_vertices, right_vertices, self.id_
        )

        # Adjacent lanes
        self._set_adjacent_lanes(lanelet)

        return lanelet

    def _calc_border_positions(self, precision: float) -> np.ndarray:
        """Determine the positions along the border where the coordinates
        of the border should be calculated.

        Args:
          precision: Number which indicates at which space interval (in curve parameter ds)
            the coordinates of the boundaries should be calculated.

        Returns:
          Array with the ordered positions.
        """
        poses = np.array([])
        for i, parametric_lane in enumerate(self.parametric_lanes):
            num_steps = int(max(2, np.ceil(parametric_lane.length / float(precision))))
            if not i:
                idx = 0
            else:
                idx = 1

            poses = np.append(
                poses,
                np.linspace(0, parametric_lane.length, num_steps)[idx::]
                + self._geo_lengths[i],
            )

        return poses

    def _set_adjacent_lanes(self, lanelet: ConversionLanelet):
        """While converting a ParametricLaneGroup to a Lanelet, set
        the proper attributes relating to adjacent lanes.

        Args:
          lanelet: The lanelet which is created from the ParametricLaneGroup.
        """
        if self.inner_neighbour is not None:
            lanelet.adj_left = self.inner_neighbour
            lanelet.adj_left_same_direction = self.inner_neighbour_same_direction

        if self.outer_neighbour is not None:
            lanelet.adj_right = self.outer_neighbour
            lanelet.adj_right_same_direction = True

    def maximum_width(self) -> float:
        """Get the maximum width of the lanelet.

        Returns:
          Maximum width of all ParametricLanes in this Group.
        """
        total_maximum = 0

        for plane in self.parametric_lanes:
            _, maximum = plane.maximum_width()
            if maximum > total_maximum:
                total_maximum = maximum

        return total_maximum

    def first_zero_width_change_position(
            self, reverse: bool = False, reference_width: float = 0.0
    ) -> float:
        """Get the earliest point of the ParametricLaneGroup where the width change is zero.

        Args:
          reverse: True if checking should start from end of lanelet.
          reference_width: Width for which width at zero width change position has
            to be greater as.

        Returns:
          Position of ParametricLaneGroup (in curve parameter ds) where width change is zero.
        """
        s_pos = 0
        positions = []

        # total_maximum = self.maximum_width()

        for plane in self.parametric_lanes:
            zero_change_positions = plane.zero_width_change_positions()
            for pos in zero_change_positions:
                positions.append((pos + s_pos, plane.calc_width(pos)))
            s_pos += plane.length

        if reverse:
            positions = list(reversed(positions))

        # if lanelet has zero width change and its width
        # is either near the maximum or it is greater than the reference width
        # the position can be used
        for pos, val in positions:
            if val > 0.9 * reference_width or val > 0.9 * self.maximum_width():
                if (pos == 0.0 and not reverse) or (pos == self.length and reverse):
                    continue
                return pos, val

        return None, None



import os
import traceback
import collections
from datetime import datetime
import numpy as np
import shapefile
import dbfread
import scipy.spatial as spt
from pyproj import Proj
from .config import default_lane_width, min_width, join_point_distance
from .config import threshold_length, threshold_distance_max, threshold_distance_min
from .functions import calculate_center, generate_random_string
from .functions import calculate_lanes_with_same_points, calculate_laneNumber_list, line2surface
from .....Tessng.MyMenu import ProgressDialog as pgd


# 获取路段信息
def get_links_info(folder_path, is_use_lon_and_lat, is_use_center_line, laneFileName="lane", proj_mode=None):
    # 文件路径
    filePath_dbf_lane = os.path.join(folder_path, f"{laneFileName}.dbf")
    filePath_shp_lane = os.path.join(folder_path, f"{laneFileName}.shp")
    
    if not os.path.exists(filePath_dbf_lane) or not os.path.exists(filePath_shp_lane):
        raise Exception("Some files are missing !")
    
    # 读取文件
    try:
        all_data_dbf_lane = list(dbfread.DBF(filePath_dbf_lane, encoding='utf-8'))
        all_data_shp_lane = shapefile.Reader(filePath_shp_lane, encoding='utf-8').shapes()
    except:
        all_data_dbf_lane = list(dbfread.DBF(filePath_dbf_lane, encoding='gbk'))
        all_data_shp_lane = shapefile.Reader(filePath_shp_lane, encoding='gbk').shapes()
    
    # 寻找中心位置
    lon_list, lat_list = [], []
    for lane_data_shp in all_data_shp_lane:
        lon_list.extend([point[0] for point in lane_data_shp.points])
        lat_list.extend([point[1] for point in lane_data_shp.points])
    if is_use_lon_and_lat:
        lon_0, lat_0 = calculate_center(min(lat_list), max(lat_list), min(lon_list), max(lon_list))
    else:
        lon_0 = (min(lon_list) + max(lon_list)) / 2
        lat_0 = (min(lat_list) + max(lat_list)) / 2

    global p
    other_info = {
        "data_source": "Shape",
        "created_time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
    }
    if is_use_lon_and_lat:
        if "prj" in proj_mode:
            filePath_prj = os.path.join(folder_path, f"{laneFileName}.prj")
            proj_string = open(filePath_prj, "r").read()
        elif "tmerc" in proj_mode:
            proj_string = f'+proj=tmerc +lon_0={lon_0} +lat_0={lat_0} +ellps=WGS84'
        elif "utm" in proj_mode:
            proj_string = f'+proj=utm +zone={lon_0 // 6 + 31} +ellps=WGS84'
        elif "web" in proj_mode:
            proj_string = 'EPSG:3857'
        else:
            proj_string = f'+proj=tmerc +lon_0={lon_0} +lat_0={lat_0} +ellps=WGS84'
            print("Warning: No this projection mode!")
        p = Proj(proj_string)
        other_info["proj_string"] = proj_string
    else:
        p = lambda x, y: (x, y)
    
    x_center, y_center = p(lon_0, lat_0)
    x_move, y_move = -float(round(x_center, 3)), -float(round(y_center, 3))
    other_info["move_distance"] = {"x_move": x_move, "y_move": y_move}

    # 保存路段信息
    links_info = collections.defaultdict(lambda: {'lanes_data': {}, 'obj': None})
    # 保存车道信息
    lanes_info = {}
    # 读取数据，并组合dbf数据和shp数据
    for index, lane_data_shp in pgd.progress(enumerate(all_data_shp_lane), "数据读取中（1/8）"):
        # 获取该路段的dbf数据
        lane_data_dbf = all_data_dbf_lane[index]
        
        # 经纬度转为平面坐标
        xy_list = [p(point[0], point[1]) for point in lane_data_shp.points]
        
        # 判断属性z是否存在
        if hasattr(lane_data_shp, "z"):
            z_list = lane_data_shp.z
        else:
            z_list = [0 for _ in range(len(xy_list))]
        
        # 获取三维坐标
        points_3D = [[x, y, z] for (x, y), z in zip(xy_list, z_list)]
        
        # 记录link车道级信息, 两种关系，路段-车道 和 车道-路段
        road_id = lane_data_dbf.get('roadId')

        # 如果没有路段ID就跳过
        if not road_id:
            print("缺失路段ID！")
            continue

        # 车道ID用生成的随机字符串
        lane_id = lane_data_dbf.get('id') or generate_random_string()
        
        lane_type = lane_data_dbf.get('type') or "driving"
        lane_width = lane_data_dbf.get('width') or default_lane_width[lane_type]

        links_info[road_id]["lanes_data"][lane_id] = {
            'roadId': road_id,
            'type': lane_type,
            'width': lane_width,
            'points': points_3D,
        }

    # 使用边界线才用到这个
    global_lane_id = 0
    for road_id in pgd.progress(list(links_info.keys()), "数据解析中（2/8）"):
        # (1) 重新计算点，保证同一路段各条车道的断点的数量相同
        lanes_points = [lane["points"] for lane in links_info[road_id]["lanes_data"].values()]
        new_lanes_points = calculate_lanes_with_same_points(lanes_points, threshold_length, threshold_distance_max, threshold_distance_min, force=False)
        if not new_lanes_points:
            print(f"不符合约束条件：{road_id}")
            del links_info[road_id]
            continue

        for index, lane_id in enumerate(links_info[road_id]["lanes_data"].keys()):
            links_info[road_id]["lanes_data"][lane_id]["points"] = new_lanes_points[index]
        
        # (2) 分中心线和边界线，获取车道的左/中/右线
        # (2.1) 如果使用中心线
        if is_use_center_line:
            # 遍历该路段上的车道
            for lane_id, lane in links_info[road_id]["lanes_data"].items():
                lane_width = lane["width"]
                lane_points_centerLine = lane["points"]
                lane_points_threeLine = line2surface(lane_points_centerLine, {
                    "right": ['right', lane_width * 0.5],
                    "center": ['right', lane_width * 0],
                    "left": ['right', lane_width * -0.5],
                })
                links_info[road_id]["lanes_data"][lane_id]["points"] = lane_points_threeLine
        # (2.2) 如果使用边界线
        else:
            temp_data_roadId = []
            temp_data_type = []
            temp_data_points = []
            
            # 遍历该路段上的车道
            for lane_id, lane in links_info[road_id]["lanes_data"].items():
                road_id = lane["roadId"]
                temp_data_roadId.append(road_id)
                
                lane_type = lane["type"]
                temp_data_type.append(lane_type)
                
                lane_points_boundaryLine = lane["points"]
                temp_data_points.append(lane_points_boundaryLine)
            
            # 清空字典，不用原来的车道ID
            links_info[road_id]["lanes_data"].clear()
            
            # 获取边界线顺序，从右向左
            order = calculate_laneNumber_list(temp_data_points)

            for i in range(1, len(temp_data_points)):
                road_id = temp_data_roadId[order.index(i)]
                lane_type = temp_data_type[order.index(i)]

                # 取中值计算中心线点位
                rightLine = temp_data_points[order.index(i)]
                leftLine = temp_data_points[order.index(i+1)]
                centerLine = [tuple(v) for v in (np.array(rightLine) + np.array(leftLine)) / 2]
                
                lane_points_threeLine = {
                    "left": leftLine,
                    "right": rightLine,
                    "center": centerLine,
                    }
                
                links_info[road_id]["lanes_data"][global_lane_id] = {
                    'roadId': road_id,
                    'type': lane_type,
                    'points': lane_points_threeLine
                    }
                
                global_lane_id += 1
        
        # (3) 添加laneNumber
        lanes_points = [lane["points"]["center"] for lane in links_info[road_id]["lanes_data"].values()]
        # 获取车道序号列表
        laneNumber_list = calculate_laneNumber_list(lanes_points)
        # 添加laneNumber信息
        for index, lane_id in enumerate(links_info[road_id]["lanes_data"].keys()):
            links_info[road_id]["lanes_data"][lane_id]["laneNumber"] = laneNumber_list[index]
            
            # 保存单独车道信息
            lanes_info[lane_id] = links_info[road_id]["lanes_data"][lane_id]
            
        # 按车道序号排序，主要是为了中心线的情况
        links_info[road_id]["lanes_data"] = dict(sorted(links_info[road_id]["lanes_data"].items(), key=lambda item:item[1]["laneNumber"]))
    
    return links_info, lanes_info, other_info


# 获取连接段信息
def get_connector_info(folder_path, lanes_info, laneConnectorFileName="laneConnector"):
    # 文件路径
    filePath_dbf_laneConnector = os.path.join(folder_path, f"{laneConnectorFileName}.dbf")
    filePath_shp_laneConnector = os.path.join(folder_path, f"{laneConnectorFileName}.shp")

    # 如果没有连接段文件
    if not os.path.exists(filePath_dbf_laneConnector) or not os.path.exists(filePath_shp_laneConnector):
        return collections.defaultdict(list)
    
    # 读取数据
    try:
        all_data_dbf_laneConnector = list(dbfread.DBF(filePath_dbf_laneConnector, encoding='utf-8'))
        all_data_shp_laneConnector = shapefile.Reader(filePath_shp_laneConnector, encoding='utf-8').shapes()
    except:
        all_data_dbf_laneConnector = list(dbfread.DBF(filePath_dbf_laneConnector, encoding='gbk'))
        all_data_shp_laneConnector = shapefile.Reader(filePath_shp_laneConnector, encoding='gbk').shapes()
    
    # 保存连接段信息
    connector_info = collections.defaultdict(list)
    
    # 读取数据，并组合dbf数据和shp数据
    for index, laneConnector_data_shp in pgd.progress(enumerate(all_data_shp_laneConnector), "数据读取中（3/8）"):
        # 获取该路段的dbf数据
        laneConnector_data_dbf = all_data_dbf_laneConnector[index]
        
        # 经纬度转为平面坐标
        xy_list = [p(point[0], point[1]) for point in laneConnector_data_shp.points]
        
        # 判断属性z是否存在
        if hasattr(laneConnector_data_shp, "z"):
            z_list = laneConnector_data_shp.z
        else:
            z_list = [0 for i in range(len(xy_list))]
        
        # 获取上下游路段ID
        try:
            from_lane_id = laneConnector_data_dbf['preLaneId']
            to_lane_id = laneConnector_data_dbf['sucLaneId']
            from_lane_info = lanes_info[from_lane_id]
            to_lane_info = lanes_info[to_lane_id]
        except:
            error = str(traceback.format_exc())
            print(error)
            continue
        
        # 车道连接的属性
        lane_type_from = from_lane_info["type"]
        lane_type_to = to_lane_info["type"]
        if lane_type_from != lane_type_to:
            continue

        lane_width = laneConnector_data_dbf.get('width') or default_lane_width.get(lane_type_from) or 3.5
        
        # 获取三维坐标
        points_3D = [[x, y, z] for (x, y), z in zip(xy_list, z_list)]
        
        lane_points_threeLine = line2surface(points_3D, {
            "right": ['right', lane_width * 0.5],
            "center": ['right', lane_width * 0],
            "left": ['right', lane_width * -0.5],
        })
        
        
        # 将同上下游路段的连接器放在一起
        connector_road_tuple = (from_lane_info['roadId'], to_lane_info['roadId'])
        connector_info[connector_road_tuple].append(
            {
                "from_lane_number": from_lane_info['laneNumber'],
                "to_lane_number": to_lane_info['laneNumber'],
                'points': lane_points_threeLine,
            }
        )

    return connector_info


# 根据距离新增连接属性
def get_new_connector_info(links_info, connector_info):
    nodes = []
    for road_id in pgd.progress(links_info, "数据解析中（4/8）"):
        for lane_info in links_info[road_id]["lanes_data"].values():
            lane_points = lane_info['points']["center"]
            nodes += [
                {
                    'point': lane_points[0],
                    'contactPoint': 'end',  # 车道起点、为连接的终点
                    'laneNumber': lane_info['laneNumber'],
                    'roadId': road_id,
                    'type': lane_info["type"]
                },
                {
                    'point': lane_points[-1],
                    'contactPoint': 'start',  # 车道终点、为连接的起点
                    'laneNumber': lane_info['laneNumber'],
                    'roadId': road_id,
                    'type': lane_info["type"]
                },
            ]

    # 寻找最近点
    node_points = [node['point'] for node in nodes]

    if node_points:
        # 用于快速查找的KDTree类
        kt = spt.KDTree(data=node_points, leafsize=10)

        # 获取距离较近的所有的点对
        node_groups = kt.query_pairs(join_point_distance)

        # 保持之前已经存在的连接段键值
        exist_connector_tuples = list(connector_info.keys())

        for node_group in pgd.progress(node_groups, "数据解析中（5/8）"):
            node_0, node_1 = nodes[node_group[0]], nodes[node_group[1]]
            # 如果link中的两个点既不在一条路段上，又不是同一起终类型，可以认为具有连接关系
            if node_0['roadId'] != node_1['roadId'] and node_0['contactPoint'] != node_1['contactPoint'] and node_0['type'] == node_1['type']:

                if node_0['contactPoint'] == 'start':
                    from_lane_info, to_lane_info = node_0, node_1
                else:
                    from_lane_info, to_lane_info = node_1, node_0
                connector_road_tuple = (from_lane_info['roadId'], to_lane_info['roadId'])
                from_lane_number, to_lane_number = from_lane_info['laneNumber'], to_lane_info['laneNumber']

                # 如果之前没有
                if connector_road_tuple not in exist_connector_tuples:
                    connector_info[connector_road_tuple].append(
                        {
                            "from_lane_number": from_lane_number,
                            "to_lane_number": to_lane_number,
                        }
                    )
    
    return connector_info


# 对拓宽段进行处理，窄的地方直接裁掉（导入边界线才用到）
def get_new_links_info(links_info):
    for road_id in pgd.progress(list(links_info.keys()), "数据解析中（6/8）"):
        # 只有该路段有至少两条车道才进行处理
        if len(links_info[road_id]["lanes_data"]) == 1:
            continue

        # 分别处理左侧拓宽和右侧拓宽的情况
        for num in [0, -1]:
            lane_id = list(links_info[road_id]["lanes_data"].keys())[num]
            lane_data = links_info[road_id]["lanes_data"][lane_id]
            leftLine = lane_data["points"]["left"]
            rightLine = lane_data["points"]["right"]
            # 只把大于最小宽度的点序号接入临时列表
            temp_list = []
            for j in range(len(leftLine)):
                left_x, left_y = leftLine[j][0], leftLine[j][1]
                right_x, right_y = rightLine[j][0], rightLine[j][1]
                distance = ((left_x - right_x) ** 2 + (left_y - right_y) ** 2) ** 0.5
                # 小于一定宽度路段要删除
                if distance >= min_width:
                    temp_list.append(j)
            # 如果有比较窄的地方
            if len(temp_list) != len(links_info[road_id]["lanes_data"][lane_id]["points"]["left"]):
                # N车道的路段
                for lane_id in links_info[road_id]["lanes_data"]:
                    for j in ["left", "center", "right"]:
                        links_info[road_id]["lanes_data"][lane_id]["points"][j] = [
                            links_info[road_id]["lanes_data"][lane_id]["points"][j][k] for k in temp_list]
    
    return links_info


def get_data(folder_path, is_use_lon_and_lat, is_use_center_line, laneFileName, laneConnectorFileName, proj_mode=None):
    # 读取路段数据
    links_info, lanes_info, other_info = get_links_info(folder_path, is_use_lon_and_lat, is_use_center_line, laneFileName, proj_mode)
    
    # 导入的中心线
    if is_use_center_line:
        # 读取连接段数据
        connector_info = get_connector_info(folder_path, lanes_info, laneConnectorFileName)
    # 导入的边界线
    else:
        connector_info = collections.defaultdict(list)
    
    # 对距离过近的自动连接
    connector_info = get_new_connector_info(links_info, connector_info)
    
    # 处理渐变段
    if not is_use_center_line:
        links_info = get_new_links_info(links_info)
    else:
        # 虚拟进度条
        for _ in pgd.progress(range(100), "数据解析中（6/8）"):
            pass
    
    return links_info, connector_info, other_info


import time
import math
from typing import Callable


def get_traj_data(simuiface, proj:Callable, p2m:Callable, move:dict):
    # 当前正在运行车辆列表
    lAllVehi = simuiface.allVehiStarted()
    # 当前已仿真时间，单位：毫秒
    simu_time = simuiface.simuTimeIntervalWithAcceMutiples()
    # 开始仿真的现实时间戳，单位：毫秒
    start_time = simuiface.startMSecsSinceEpoch()
    # 当前仿真计算批次
    batchNum = simuiface.batchNumber()

    traj_data = {
        "timestamp": str(int(time.time() * 1000)),
        "simuTime": simu_time,
        'startSimuTime': start_time,
        "batchNum": batchNum,
        "count": len(lAllVehi),
        "objs": [],
    }
    
    for vehi in lAllVehi:
        x = p2m(vehi.pos().x()) + move["x_move"]
        y = -p2m(vehi.pos().y()) + move["y_move"]
        if math.isnan(x) or math.isnan(y):
            continue

        lon, lat = proj(x, y, inverse=True)

        in_link = vehi.roadIsLink()
        lane = vehi.lane()

        # 车辆寻找异常，跳过
        if (in_link and not vehi.lane()) or (not in_link and not vehi.laneConnector()):
            continue

        angle = vehi.angle()
        veh_data = {
            "id": vehi.id(),
            'typeCode': vehi.vehicleTypeCode(),
            'roadId': vehi.roadId(),
            'inLink': in_link,
            'laneCount': in_link and lane.link().laneCount(),
            'laneNumber': in_link and lane.number(),
            'laneTypeName': in_link and lane.actionType(),
            'angle': angle,
            'speed': p2m(vehi.currSpeed()),
            'Speed': p2m(vehi.currSpeed()) * 3.6,
            'size': [p2m(vehi.length()), 2, 2],
            'color': "",
            'x': x,
            'y': y,
            'z': vehi.v3z(),
            'longitude': lon,
            'latitude': lat,
            'eulerX': -angle / 180 * math.pi + math.pi / 2,
            'eulerY': -angle / 180 * math.pi + math.pi / 2,
            'eulerZ': -angle / 180 * math.pi + math.pi / 2,
        }

        traj_data['objs'].append(veh_data)

    return traj_data




import collections
import numpy as np
from typing import Dict, List
from PySide2.QtGui import QVector3D

from .config import *


def p2m(x):
    return x * sceneScale

def m2p(x):
    return x / sceneScale


def get_section_childs(section_info: Dict, lengths: List, direction: str) -> List[Dict]:
    """
        处理某 Section 某方向的路网信息
        根据二维信息以及自定义配置，将section 进行再次处理，将宽度过窄的车道所在的路段打断，分配合适的连接段或者抹除车道
    Args:
        section_info:
        lengths: 因为分段时，以road作为最小单位，所以需要提供此 section 对应的长度断点序列
        direction: 沿参考线方向

    Returns:
        被处理后的 小路段列表
    """
    # 分为左右车道，同时过滤tess规则下不同的车道类型
    if direction == 'left':
        lane_ids = [lane_id for lane_id in section_info['lanes'].keys() if
                    lane_id > 0 and lane_id in section_info["tess_lane_ids"]]
    else:
        lane_ids = [lane_id for lane_id in section_info['lanes'].keys() if
                    lane_id < 0 and lane_id in section_info["tess_lane_ids"]]
    # 路段遍历，获取link段，处理异常点
    point_infos = []
    # 查找连接段
    for index, length in enumerate(lengths):
        point_info = {
            'lanes': {},
            'is_link': True,
        }
        for lane_id in lane_ids:
            lane_info = section_info['lanes'][lane_id]
            tess_lane_type = LANE_TYPE_MAPPING.get(lane_info['type'])
            if tess_lane_type not in WIDTH_LIMIT.keys() or lane_info['widths'][index] > \
                    WIDTH_LIMIT[tess_lane_type]['split']:
                point_info['lanes'][lane_id] = lane_info['type']  # 无宽度限制或者宽度足够，正常车道
            elif lane_info['widths'][index] > WIDTH_LIMIT[tess_lane_type]['join']:
                point_info['lanes'][lane_id] = lane_info['type']  # 宽度介于中间，作为连接段
                point_info['is_link'] = False
            # 否则，不加入车道列表
        point_infos.append(point_info)

    # 连续多个点的信息完全一致，可作为同一路段
    childs = []
    child_point = []
    start_index = None  # 分片时，start_index 为None 会视为0
    for index in range(len(lengths)):
        if len(child_point) == 1:
            start_index = index - 1
        point_info = point_infos[index]
        if index < POINT_REQUIRE:  # 首尾必须为link
            child_point.append(point_infos[0])
        elif len(lengths) - index - 1 < POINT_REQUIRE:
            child_point.append(point_infos[-1])
        elif not child_point:  # 原列表为空
            if point_info['is_link']:
                child_point.append(point_info)
            else:
                continue
        else:  # 原列表不为空
            if point_info == child_point[0]:
                child_point.append(point_info)
            elif len(child_point) >= POINT_REQUIRE:
                childs.append(
                    {
                        'start': start_index,
                        'end': index,
                        'lanes': set(child_point[0]['lanes'].keys()) & set(child_point[0]['lanes'].keys()),
                    }
                )
                child_point = []
            else:
                continue

    # 把最后一个存在的点序列导入, 最后一个应该以末尾点为准,但是如果此处包含了首&尾，应该取数据量的交集
    # 这样可能会丢失部分路段，所以在建立连接段时，必须确保 from_lane, to_lane 均存在
    childs.append(
        {
            'start': start_index,
            'end': len(lengths) - 1,
            'lanes': set(child_point[0]['lanes'].keys()) & set(child_point[0]['lanes'].keys())
        }
    )
    # lengths 只是断点序列，标记着与起始点的距离,反向用了同样的lengths
    # 得到link的列表,因为lane的点坐标与方向有关，所以此时的child已经根据方向排序
    return childs


def get_inter(string: str, roads_info: Dict) -> List:
    """
        根据 opendrive2lanelet 中定义的的车道名判断此车道的合法性，并返回相应的的 road_id,section_id,lane_id, -1
    Args:
        string: 车道名
        roads_info: 路段信息字典

    Returns:
        车道名是否合法，车道名的详细信息
    """
    inter_list = []
    is_true = True
    for i in string.split('.'):
        try:
            inter_list.append(int(i))
        except:
            inter_list.append(None)
            is_true = False

    # 检查 前后续路段 是否存在,不存在可以忽略
    if inter_list[0] not in roads_info.keys():
        is_true = False
    return [is_true, *inter_list]


def connect_childs(links: "Tessng.ILink", connector_mapping: Dict) -> None:
    """
        因为在section内对过窄车道进行了优化处理，所以可能将section切分成了多个link，需要为每个link在内部建立连接关系
    Args:
        links: section内 同方向子link 顺序列表
        connector_mapping: 全局的路段连接关系表

    Returns:
        无返回值，但路段连接关系表被扩充
    """
    for index in range(len(links) - 1):
        from_link_info = links[index]
        to_link_info = links[index + 1]
        from_link = from_link_info['link']
        to_link = to_link_info['link']
        if not (from_link and to_link and from_link_info['lane_ids'] and to_link_info['lane_ids']):
            continue

        actionTypeMapping = collections.defaultdict(lambda: {"from": set(), 'to': set()})
        for lane_id in from_link_info['lane_ids']:
            actionTypeMapping[from_link_info[lane_id].actionType()]['from'].add(lane_id)
        for lane_id in to_link_info['lane_ids']:
            actionTypeMapping[to_link_info[lane_id].actionType()]['to'].add(lane_id)

        connect_lanes = set()
        for actionType, lanes_info in actionTypeMapping.items():
            for lane_id in set(lanes_info['from'] | lanes_info['to']):
                # 车道原始编号相等，取原始编号对应的车道，否则，取临近车道(lane_ids 是有序的，所以临近车道永远偏向左边区)
                # 因为进行过自优化(人为打断与合并)，可能会导致前后失去连接关系，需要注意
                from_lane_id = min(lanes_info['from'], key=lambda x: abs(x - lane_id)) if lanes_info[
                    'from'] else None
                to_lane_id = min(lanes_info['to'], key=lambda x: abs(x - lane_id)) if lanes_info['to'] else None

                from_lane = from_link_info.get(from_lane_id)
                to_lane = to_link_info.get(to_lane_id)
                if from_lane and to_lane:
                    connect_lanes.add((from_lane.number() + 1, to_lane.number() + 1))

        if connect_lanes:
            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lFromLaneNumber'] += [i[0] for i in
                                                                                         connect_lanes]
            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lToLaneNumber'] += [i[1] for i in connect_lanes]
            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lanesWithPoints3'] += [None for _ in
                                                                                          connect_lanes]
            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['infos'] += []


# 计算两向量间的夹角
def cal_angle_of_vector(v0, v1, is_use_deg=True):
    dot_product = np.dot(v0, v1)
    v0_len = np.linalg.norm(v0)
    v1_len = np.linalg.norm(v1)
    try:
        angle_rad = np.arccos(dot_product / (v0_len * v1_len))
    except ZeroDivisionError as error:
        return None

    if is_use_deg:
        return np.rad2deg(angle_rad)
    return angle_rad


# 根据前后两点坐标绘制向量
def get_vector(start_point, end_point):
    # return np.array([end_point[index] - start_point[index] for index in range(3)])
    return np.array(end_point) - np.array(start_point)


# 根据角度限制整合中心点，减少绘制路网时的点数
def get_new_point_indexs(center_points, angle):
    new_point_indexs = [0, 1]  # 前两个点是基础
    for index, point in enumerate(center_points[2:-1]):
        real_index = index + 2  # 在 center_points 中的真实索引

        point_0 = center_points[new_point_indexs[-2]]
        point_1 = center_points[new_point_indexs[-1]]
        point_2 = point

        vector_1 = get_vector(point_0, point_1)
        vector_2 = get_vector(point_0, point_2)

        included_angle = cal_angle_of_vector(vector_1, vector_2)
        point_distance = np.sqrt(np.sum((np.array(point_2) - np.array(point_1)) ** 2))

        # 如果两向量夹角变化不大，抹除观测点(不能用连续的两个点比较,即 01 对比 12，容易对缓慢变化的路段判断错误)
        # print(included_angle, angle, point_distance >= max_length, included_angle > 0)
        # print(point_0, point_1, point_2, vector_1, vector_2)
        if included_angle is None or included_angle >= angle:
            new_point_indexs.append(real_index)
        # 如果观测点距离上一点过远，且存在角度差异，也不会抹除，不加这个可能会把微小的角度误差放大到明显的地步
        elif point_distance >= max_length and included_angle > 0:
        # elif point_distance >= max_length:  # 存在微小角度偏差的路段 可能会漫延很长
            new_point_indexs.append(real_index)
        else:
            # print('不添加')
            pass
    # 最后一个点需添加
    if len(center_points) - 1 > new_point_indexs[-1]:
        new_point_indexs.append(len(center_points) - 1)

    # TODO 尝试缩减第二个点
    def is_remove_index_1(center_points, new_point_indexs):
        if len(new_point_indexs) < 3:
            # 只有两个点，无法缩减
            return False
        point_0 = center_points[new_point_indexs[0]]
        point_1 = center_points[new_point_indexs[1]]
        point_2 = center_points[new_point_indexs[2]]

        vector_1 = get_vector(point_0, point_1)
        vector_2 = get_vector(point_0, point_2)

        included_angle = cal_angle_of_vector(vector_1, vector_2)

        # 距离取较小值
        point_distance = np.sqrt(np.sum((np.array(point_1) - np.array(point_0)) ** 2))

        # 如果两向量夹角变化不大，抹除观测点(不能用连续的两个点比较,即 01 对比 12，容易对缓慢变化的路段判断错误)
        if included_angle is None or included_angle >= angle:
            return False
        # 如果观测点距离上一点过远，且存在角度差异，也不会抹除，不加这个可能会把微小的角度误差放大到明显的地步
        elif point_distance >= max_length and included_angle > 0:
            # elif point_distance >= max_length:  # 存在微小角度偏差的路段 可能会蔓延很长
            return False
        else:
            return True

    if is_remove_index_1(center_points, new_point_indexs):
        del new_point_indexs[1]
    # print(f"point 缩减 {len(center_points)} to {len(new_point_indexs)}")
    return new_point_indexs



# =====================================
# 处理过短连接段

def qtpoint2point(qt_points):
    points = []
    for qt_point in qt_points:
        x = p2m(qt_point.x())
        y = -p2m(qt_point.y())
        z = p2m(qt_point.z())
        points.append([round(x,2), round(y,2), round(z,2)])
    return points


def get_coo_list(vertices):
    list1 = []
    for index in range(0, len(vertices), 1):
        vertice = vertices[index]
        # list1.append(QPointF(m2p((vertice[0])), m2p(-(vertice[1]))))
        list1.append(QVector3D(m2p((vertice[0])), m2p(-(vertice[1])), m2p(vertice[2])))
    if len(list1) < 2:
        raise 3
    return list1


def calculate_section_length(p1, p2):
    dp = np.array(p1) - np.array(p2)
    return np.linalg.norm(dp[:2])

def interpolate_point_on_segment(p1, p2, t):
    return p1 + t * (p2 - p1)

def split_line(line, given_distance, mode):
    total_length = np.sum([calculate_section_length(line[i - 1], line[i]) for i in range(1, len(line))])

    if given_distance >= total_length:
        print(f"Warning: given_distance {given_distance} is greater than the line's total length {total_length}.")

    if mode == "start":
        given_distance = min(given_distance, total_length*0.2)
    elif mode == "end":
        given_distance = max(total_length - given_distance, total_length*0.8)

    all_length = 0
    for i in range(1, len(line)):
        p1 = np.array(line[i - 1])
        p2 = np.array(line[i])
        section_length = calculate_section_length(p1, p2)
        all_length += section_length

        if all_length < given_distance:
            continue
        elif all_length == given_distance:
            line1 = line[:i + 1]
            line2 = line[i:]
            info = (i, None)
            break
        else:
            before_length = all_length - section_length
            ratio = round((given_distance - before_length) / section_length, 3)
            cut_point = interpolate_point_on_segment(p1, p2, ratio)
            line1 = line[:i] + [cut_point.tolist()]
            line2 = [cut_point.tolist()] + line[i:]
            info = (i-1, ratio)
            break

    if mode == "start":
        line = line2
    elif mode == "end":
        line = line1
    return line, info

def split_line_2(line, info, mode):
    index, ratio = info

    if ratio is None:
        line1 = line[:index + 1]
        line2 = line[index:]
    else:
        p1 = np.array(line[index])
        p2 = np.array(line[index+1])
        cut_point = interpolate_point_on_segment(p1, p2, ratio)
        line1 = line[:index+1] + [cut_point.tolist()]
        line2 = [cut_point.tolist()] + line[index+1:]

    if mode == "start":
        return_line = line2
    elif mode == "end":
        return_line = line1
    return return_line


# 处理过短连接段
# =====================================



from PySide2.QtGui import QVector3D
from .config import sceneScale


def p2m(x):
    return x * sceneScale

def qtpoint2point(qtpoints, move=None):
    if move is None:
        x_move, y_move = 0, 0
    else:
        x_move, y_move = -move["x_move"], -move["y_move"]

    points = []
    for qt_point in qtpoints:
        x = p2m(qt_point.x())
        y = -p2m(qt_point.y())
        z = p2m(qt_point.z())
        x += x_move
        y += y_move
        points.append([x, y, z])
    return points

# 坐标转经纬度
def point2latlon(points, p):
    new_points = []
    for point in points:
        lon, lat = p(point[0], point[1], inverse=True)
        new_points.append([lon, lat, point[2]])
    # 返回经纬度坐标及z坐标列表
    return new_points





# 场景比例尺
sceneScale = 1

# unity 属性映射关系
convert_attribute_mapping = {
    "black": "Driving",
    "white": "WhiteLine",
    "yellow": "YellowLine",
}

filter_ids = []
border_line_width = 0.2
center_line_width = 0.3
empty_line_lenfth, real_line_length = 3, 4  # 虚实线长度

# # unity 信息提取的类型映射
# UNITY_LANE_MAPPING = {
#     "Driving": ["driving", "stop", "parking", "entry", "exit", "offRamp", "onRamp", "connectingRamp", ],
#     "None": ["none"],
#     "GreenBelt": ["shoulder", "border", "median", "curb"],
#     "SideWalk": ["sidewalk"],
#     "Biking": ["biking", ],
#     "Restricted": ["restricted"],
#     "WhiteLine": [],
#     "YellowLine": [],
#     "Other": ["bidirectional", "special1", "special2", "special3", "roadWorks", "tram", "rail", ]
# }

# 为了节省空间，仅对路段级的做黑色三角形，同时取车道左侧做白色线，一般为白色虚线，如果车道类型不一致，做白色实线，如果是最左侧车道，做黄色实线



# -*- coding: utf-8 -*-

__author__ = "Benjamin Orthen, Stefan Urban"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


class OpenDrive:
    """ """

    def __init__(self):
        self.header = None
        self._roads = []
        self._controllers = []
        self._junctions = []
        self._junctionGroups = []
        self._stations = []

    # @property
    # def header(self):
    #     return self._header

    @property
    def roads(self):
        """ """
        return self._roads

    def getRoad(self, id_):
        """

        Args:
          id_:

        Returns:

        """
        for road in self._roads:
            if road.id == id_:
                return road

        return None

    @property
    def controllers(self):
        """ """
        return self._controllers

    @property
    def junctions(self):
        """ """
        return self._junctions

    def getJunction(self, junctionId):
        """

        Args:
          junctionId:

        Returns:

        """
        for junction in self._junctions:
            if junction.id == junctionId:
                return junction
        return None

    @property
    def junctionGroups(self):
        """ """
        return self._junctionGroups

    @property
    def stations(self):
        """ """
        return self._stations


class Header:
    """ """

    def __init__(
            self,
            rev_major=None,
            rev_minor=None,
            name: str = None,
            version=None,
            date=None,
            north=None,
            south=None,
            east=None,
            west=None,
            vendor=None,
    ):
        self.revMajor = rev_major
        self.revMinor = rev_minor
        self.name = name
        self.version = version
        self.date = date
        self.north = north
        self.south = south
        self.east = east
        self.west = west
        self.vendor = vendor
        self.geo_reference = None



from .utils import config


def excel2tess(netiface, params):
    # params:
    # - file_path
    
    file_path = params["file_path"]
    
    config.sceneScale = netiface.sceneScale()
    
    from .utils.get_data import get_data
    from .utils.create_network import create_network
    
    # 读取数据
    links_data, other_data = get_data(file_path) # dict, dict

    # 创建路段和连接段
    if links_data:
        create_network(netiface, links_data, other_data)
        status, message = True, None
    else:
        status, message = False, "所选文件中无数据或无合法数据"

    return status, message





import json
from pyproj import Proj
from PySide2.QtCore import QObject, Signal

from ..DLLs.Tessng import PyCustomerSimulator, tessngIFace, p2m
from ..Toolbox.traj_output.get_traj_data import get_traj_data
from ..Toolbox.traj_output.send_traj_data import SendData
from ..Tessng.MyMenu import GlobalVar


class MySimulator(QObject, PyCustomerSimulator):
    signalRunInfo = Signal(str)
    forStopSimu = Signal()
    forReStartSimu = Signal()

    def __init__(self):
        QObject.__init__(self)
        PyCustomerSimulator.__init__(self)

        # 投影关系
        self.traj_proj = None
        # move
        self.traj_move = None
        # 数据发送线程
        self.send_data_class = None

    # 仿真开始前
    def beforeStart(self, ref_keepOn):
        traj_proj_string = GlobalVar.traj_proj_string # str
        traj_json_config = GlobalVar.traj_json_config # str
        traj_kafka_config = GlobalVar.traj_kafka_config # dict

        # 投影关系
        if GlobalVar.traj_proj_string:
            self.traj_proj = Proj(traj_proj_string)
        else:
            self.traj_proj = lambda x,y,inverse=None: (None,None)

        # move
        iface = tessngIFace()
        netiface = iface.netInterface()
        move = netiface.netAttrs().otherAttrs().get("move_distance")
        if move is None or "tmerc" in traj_proj_string:
            self.traj_move = {"x_move": 0, "y_move": 0}
        else:
            self.traj_move = {"x_move": -move["x_move"], "y_move": -move["y_move"]}

        # 数据发送线程
        if traj_json_config or traj_kafka_config:
            self.send_data_class = SendData(traj_json_config, traj_kafka_config)

    # 仿真结束后
    def afterStop(self):
        # 投影关系
        self.traj_proj = None

        # move
        self.traj_move = None

        # 数据发送进程
        if self.send_data_class is not None:
            self.send_data_class.close()
            self.send_data_class = None

    # 每帧仿真后
    def afterOneStep(self):
        iface = tessngIFace()
        simuiface = iface.simuInterface()

        # 如果不需要导出，就不需要计算
        if self.send_data_class is None:
            return

        # 计算轨迹数据
        traj_data = get_traj_data(simuiface, self.traj_proj, p2m, self.traj_move)

        # 发送轨迹数据
        self.send_data_class.put_data(traj_data)




# -*- coding: utf-8 -*-


"""Module to contain Network which can load an opendrive object and then export
to lanelets. Iternally, the road network is represented by ParametricLanes."""

from commonroad.scenario.scenario import Scenario
from .opendriveparser.elements.opendrive import OpenDrive
from .utils import encode_road_section_lane_width_id
from .conversion_lanelet_network import ConversionLaneletNetwork
from .converter import OpenDriveConverter
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


__author__ = "Benjamin Orthen, Stefan Urban"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


class Network:
    """Represents a network of parametric lanes, with a LinkIndex
    which stores the neighbor relations between the parametric lanes.

    Args:

    """

    def __init__(self):
        self._planes = []
        self._link_index = None

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def load_opendrive(self, opendrive: OpenDrive):
        """Load all elements of an OpenDRIVE network to a parametric lane representation

        Args:
          opendrive:

        """

        if not isinstance(opendrive, OpenDrive):
            raise TypeError()

        self._link_index = LinkIndex()
        self._link_index.create_from_opendrive(opendrive)

        # Convert all parts of a road to parametric lanes (planes)
        for road in pgd.progress(opendrive.roads, '数据解析中（2/13）'):
            road.planView.precalculate()

            # The reference border is the base line for the whole road
            reference_border = OpenDriveConverter.create_reference_border(
                road.planView, road.lanes.laneOffsets
            )

            # A lane section is the smallest part that can be converted at once
            for lane_section in road.lanes.lane_sections:
                parametric_lane_groups = OpenDriveConverter.lane_section_to_parametric_lanes(
                    lane_section, reference_border
                )

                self._planes.extend(parametric_lane_groups)

    def export_lanelet_network(
            self, filter_types: list = None, roads_info=None, context=None
    ) -> "ConversionLaneletNetwork":
        """Export network as lanelet network.

        Args:
          filter_types: types of ParametricLane objects to be filtered. (Default value = None)

        Returns:
          The converted LaneletNetwork object.
        """

        # Convert groups to lanelets
        lanelet_network = ConversionLaneletNetwork()
        progress = 0
        for parametric_lane in pgd.progress(self._planes, '数据解析中（3/13）'):
            if filter_types is not None and parametric_lane.type not in filter_types:
                continue
            road_id = int(parametric_lane.id_.split('.')[0])
            section_id = int(parametric_lane.id_.split('.')[1])
            lengths = roads_info[road_id]['road_points'][section_id]['lengths']

            # 传入的是整条参考线点序列，需要减去初始值得到section对应的lane的点序列,length只是长度序列，对左右来向是一样的，不会影响计算
            length_start = lengths[0]
            poses = [min(i - length_start, parametric_lane.length) for i in lengths]

            lanelet = parametric_lane.to_lanelet(0.5, poses)
            lanelet.type = parametric_lane.type

            lanelet.predecessor = self._link_index.get_predecessors(parametric_lane.id_)
            lanelet.successor = self._link_index.get_successors(parametric_lane.id_)

            lanelet_network.add_lanelet(lanelet)
            progress += 1
            # send_signal(context, int(progress / len(self._planes) * 80))

        return lanelet_network

    def export_commonroad_scenario(
            self, dt: float = 0.1, benchmark_id=None, filter_types=None, roads_info=None, context=None
    ):
        """Export a full CommonRoad scenario

        Args:
          dt:  (Default value = 0.1)
          benchmark_id:  (Default value = None)
          filter_types:  (Default value = None)

        Returns:

        """
        if benchmark_id is None:
            benchmark_id = "none"
            
        # print(dt, benchmark_id)
        
        scenario = Scenario(dt=dt, benchmark_id=benchmark_id)

        scenario.add_objects(
            self.export_lanelet_network(
                filter_types=filter_types
                if isinstance(filter_types, list)
                else ["driving", "onRamp", "offRamp", "exit", "entry"],
                roads_info=roads_info,
                context=context,
            )
        )

        return scenario


class LinkIndex:
    """Overall index of all links in the file, save everything as successors, predecessors can be found via a reverse search"""

    def __init__(self):
        self._successors = {}

    def create_from_opendrive(self, opendrive):
        """Create a LinkIndex from an OpenDrive object.

        Args:
          opendrive: OpenDrive style object.

        Returns:

        """
        self._add_junctions(opendrive)
        # Extract link information from road lanes
        for road in opendrive.roads:
            for lane_section in road.lanes.lane_sections:
                for lane in lane_section.allLanes:
                    parametric_lane_id = encode_road_section_lane_width_id(
                        road.id, lane_section.idx, lane.id, -1
                    )

                    # Not the last lane section? > Next lane section in same road
                    if lane_section.idx < road.lanes.getLastLaneSectionIdx():
                        successorId = encode_road_section_lane_width_id(
                            road.id, lane_section.idx + 1, lane.link.successorId, -1
                        )

                        self.add_link(parametric_lane_id, successorId, lane.id >= 0)

                    # Last lane section! > Next road in first lane section
                    # Try to get next road
                    elif (
                            road.link.successor is not None
                            and road.link.successor.elementType != "junction"
                    ):

                        next_road = opendrive.getRoad(road.link.successor.element_id)

                        if next_road is not None:

                            if road.link.successor.contactPoint == "start":
                                successorId = encode_road_section_lane_width_id(
                                    next_road.id, 0, lane.link.successorId, -1
                                )

                            else:  # contact point = end
                                successorId = encode_road_section_lane_width_id(
                                    next_road.id,
                                    next_road.lanes.getLastLaneSectionIdx(),
                                    lane.link.successorId,
                                    -1,
                                )
                            self.add_link(parametric_lane_id, successorId, lane.id >= 0)

                    # Not first lane section? > Previous lane section in same road
                    if lane_section.idx > 0:
                        predecessorId = encode_road_section_lane_width_id(
                            road.id, lane_section.idx - 1, lane.link.predecessorId, -1
                        )

                        self.add_link(predecessorId, parametric_lane_id, lane.id >= 0)

                    # First lane section! > Previous road
                    # Try to get previous road
                    elif (
                            road.link.predecessor is not None
                            and road.link.predecessor.elementType != "junction"
                    ):

                        prevRoad = opendrive.getRoad(road.link.predecessor.element_id)

                        if prevRoad is not None:

                            if road.link.predecessor.contactPoint == "start":
                                predecessorId = encode_road_section_lane_width_id(
                                    prevRoad.id, 0, lane.link.predecessorId, -1
                                )

                            else:  # contact point = end
                                predecessorId = encode_road_section_lane_width_id(
                                    prevRoad.id,
                                    prevRoad.lanes.getLastLaneSectionIdx(),
                                    lane.link.predecessorId,
                                    -1,
                                )
                            self.add_link(
                                predecessorId, parametric_lane_id, lane.id >= 0
                            )

    def add_link(self, parametric_lane_id, successor, reverse: bool = False):
        """

        Args:
          parametric_lane_id:
          successor:
          reverse:  (Default value = False)

        Returns:

        """

        # if reverse, call function recursively with switched parameters
        if reverse:
            self.add_link(successor, parametric_lane_id)
            return

        if parametric_lane_id not in self._successors:
            self._successors[parametric_lane_id] = []

        if successor not in self._successors[parametric_lane_id]:
            self._successors[parametric_lane_id].append(successor)

    def _add_junctions(self, opendrive):
        """

        Args:
          opendrive:

        Returns:

        """
        # add junctions to link_index
        # if contact_point is start, and laneId from connecting_road is negative
        # the connecting_road is the successor
        # if contact_point is start, and laneId from connecting_road is positive
        # the connecting_road is the predecessor
        # for contact_point == end it's exactly the other way
        for junction in opendrive.junctions:
            for connection in junction.connections:
                incoming_road = opendrive.getRoad(connection.incomingRoad)
                connecting_road = opendrive.getRoad(connection.connectingRoad)
                contact_point = connection.contactPoint
                if not (incoming_road and connecting_road):
                    continue

                for lane_link in connection.laneLinks:
                    if contact_point == "start":

                        # decide which lane section to use (first or last)
                        if lane_link.fromId < 0:
                            lane_section_idx = (
                                incoming_road.lanes.getLastLaneSectionIdx()
                            )
                        else:
                            lane_section_idx = 0
                        incoming_road_id = encode_road_section_lane_width_id(
                            incoming_road.id, lane_section_idx, lane_link.fromId, -1
                        )
                        connecting_road_id = encode_road_section_lane_width_id(
                            connecting_road.id, 0, lane_link.toId, -1
                        )
                        self.add_link(
                            incoming_road_id, connecting_road_id, lane_link.toId > 0
                        )
                    else:
                        # decide which lane section to use (first or last)
                        if lane_link.fromId < 0:
                            lane_section_idx = 0

                        else:
                            lane_section_idx = (
                                incoming_road.lanes.getLastLaneSectionIdx()
                            )
                        incoming_road_id = encode_road_section_lane_width_id(
                            incoming_road.id, 0, lane_link.fromId, -1
                        )
                        connecting_road_id = encode_road_section_lane_width_id(
                            connecting_road.id,
                            connecting_road.lanes.getLastLaneSectionIdx(),
                            lane_link.toId,
                            -1,
                        )
                        self.add_link(
                            incoming_road_id, connecting_road_id, lane_link.toId < 0
                        )

    def remove(self, parametric_lane_id):
        """

        Args:
          parametric_lane_id:

        """
        # Delete key
        if parametric_lane_id in self._successors:
            del self._successors[parametric_lane_id]

        # Delete all occurances in successor lists
        for successorsId, successors in self._successors.items():
            if parametric_lane_id in successors:
                self._successors[successorsId].remove(parametric_lane_id)

    def get_successors(self, parametric_lane_id: str) -> list:
        """

        Args:
          parametric_lane_id: Id of ParametricLane for which to search
            successors.

        Returns:
          List of successors belonging the the ParametricLane.
        Par
        """
        if parametric_lane_id not in self._successors:
            return []

        return self._successors[parametric_lane_id]

    def get_predecessors(self, parametric_lane_id: str) -> list:
        """

        Args:
          parametric_lane_id: Id of ParametricLane for which to search predecessors.

        Returns:
          List of predecessors of a ParametricLane.
        """
        predecessors = []
        for successor_plane_id, successors in self._successors.items():
            if parametric_lane_id not in successors:
                continue

            if successor_plane_id in predecessors:
                continue

            predecessors.append(successor_plane_id)

        return predecessors



# -*- coding: utf-8 -*-

"""Module for logic behind converting OpenDrive to ParametricLanes."""

from typing import Tuple, List
from .plane_elements.plane import (
    ParametricLane,
    ParametricLaneBorderGroup,
)
from .plane_elements.plane_group import ParametricLaneGroup
from .plane_elements.border import Border
from .utils import encode_road_section_lane_width_id

__author__ = "Benjamin Orthen"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


class OpenDriveConverter:
    """Class for static methods to convert lane_sections to parametric_lanes."""

    @staticmethod
    def create_reference_border(plan_view, lane_offsets) -> Border:
        """Create the most inner border from a PlanView.
        This border is used as a reference for other
        borders which rely on the PlanView.

        Args:
          plan_view: PlanView object from OpenDrive which specifies the geometry
            of the reference path.
          lane_offsets: Object which contains information about width offset of reference
            path the plain_view path.

        Returns:
           The reference border on which all other borders in this lane section are based upon.

        """

        reference_border = Border()

        # Set reference to plan view
        reference_border.reference = plan_view

        # Lane offsets will be coeffs
        # this has to be done if the reference path has the laneoffset attribute
        # and thus is different to the geometry described in the plan_view
        # openDRIVE lets multiple laneOffsets start at the same position
        # but only the last one counts -> delete all previous ones
        if any(lane_offsets):
            for lane_offset in lane_offsets:
                if lane_offset.start_pos in reference_border.width_coefficient_offsets:
                    # offset is already there, delete previous entries
                    idx = reference_border.width_coefficient_offsets.index(
                        lane_offset.start_pos
                    )
                    del reference_border.width_coefficient_offsets[idx]
                    del reference_border.width_coefficients[idx]
                reference_border.width_coefficient_offsets.append(lane_offset.start_pos)
                reference_border.width_coefficients.append(
                    lane_offset.polynomial_coefficients
                )
        else:
            reference_border.width_coefficient_offsets.append(0.0)
            reference_border.width_coefficients.append([0.0])

        return reference_border

    @staticmethod
    def lane_section_to_parametric_lanes(
            lane_section, reference_border
    ) -> List[ParametricLaneGroup]:
        """Convert a whole lane section into a list of ParametricLane objects.

        Args:
          lane_section:
          reference_border:

        Returns:

        """

        plane_groups = []

        for side in ["right", "left"]:

            # lanes loaded by opendriveparser are aleady sorted by id
            # coeff_factor decides if border is left or right of the reference line
            lanes = (
                lane_section.rightLanes if side == "right" else lane_section.leftLanes
            )
            coeff_factor = -1.0 if side == "right" else 1.0

            # Most inner border gets added first
            lane_borders = [reference_border]

            # copy reference border, but set refOffset to start of lane_section

            for lane in lanes:

                inner_neighbour_id, outer_neighbour_id, inner_neighbour_same_dir = OpenDriveConverter.determine_neighbours(
                    lane
                )

                # Create outer lane border
                # outer_parametric_lane_border =

                lane_borders.append(
                    OpenDriveConverter._create_outer_lane_border(
                        lane_borders, lane, coeff_factor
                    )
                )

                plane_group = ParametricLaneGroup(
                    id_=encode_road_section_lane_width_id(
                        lane_section.parentRoad.id, lane_section.idx, lane.id, -1
                    ),
                    inner_neighbour=inner_neighbour_id,
                    inner_neighbour_same_direction=inner_neighbour_same_dir,
                    outer_neighbour=outer_neighbour_id,
                )

                # Create new lane for each width segment
                for width in lane.widths:
                    parametric_lane = OpenDriveConverter.create_parametric_lane(
                        lane_borders, width, lane
                    )
                    parametric_lane.reverse = bool(lane.id > 0)
                    plane_group.append(parametric_lane)

                # if lane borders are specified by offsets instead of widths
                # for borders in lane.borders:

                if plane_group.length > 0:
                    plane_groups.append(plane_group)

        return plane_groups

    @staticmethod
    def create_parametric_lane(lane_borders, width, lane) -> ParametricLane:
        """Create a parametric lane for a certain width section.

        Args:
          lane_borders: Array with already created lane borders.
          width: Width section with offset and coefficient information.
          lane: Lane in which new parametric lane is created.
          prev_inner_neighbours: Inner neighbours of parametric lane.

        Returns:
          A ParametricLane object with specified borders and a unique id.
        """

        border_group = ParametricLaneBorderGroup(
            inner_border=lane_borders[-2],
            outer_border=lane_borders[-1],
            inner_border_offset=width.start_offset + lane_borders[-1].ref_offset,
            outer_border_offset=width.start_offset,
        )
        parametric_lane = ParametricLane(
            id_=encode_road_section_lane_width_id(
                lane.lane_section.parentRoad.id,
                lane.lane_section.idx,
                lane.id,
                width.idx,
            ),
            type_=lane.type,
            length=width.length,
            border_group=border_group,
        )
        return parametric_lane

    @staticmethod
    def _create_outer_lane_border(lane_borders, lane, coeff_factor) -> Border:
        """Create an outer lane border of a lane.
        InnerBorder is already saved in lane_borders, as it is
        the outer border of the inner neighbour of the lane.

        Args:
          lane_borders: Previous calculated lane borders of more inner lanes.
          lane: Lane for which outer border shall be created.
            This is specified in parameter ds of curve length.
          coeff_factor: factor of -1 or 1, dependent on which side of the reference
            path the lane is (right side is -1).

        Returns:
          The created outer lane border.

        """
        # Create outer lane border
        # Offset from reference border is already included in first inner border
        # (lane_border[0])
        # reference_border starts at beginning of road, prev: lane section
        border = Border()
        if len(lane_borders) == 1:
            border.ref_offset = lane.lane_section.sPos

        # last created border
        if lane.has_border_record:
            border.reference = lane_borders[0]
        else:
            border.reference = lane_borders[-1]

        for width in lane.widths:
            border.width_coefficient_offsets.append(width.start_offset)
            border.width_coefficients.append(
                [x * coeff_factor for x in width.polynomial_coefficients]
            )
        return border

    @staticmethod
    def determine_neighbours(lane) -> Tuple[str, str, bool]:
        """

        Args:
          lane:

        Returns:

        """
        if abs(lane.id) > 1:

            if lane.id > 0:
                inner_lane_id = lane.id - 1
                outer_lane_id = lane.id + 1
            else:
                inner_lane_id = lane.id + 1
                outer_lane_id = lane.id - 1

            inner_neighbour_same_dir = True

        else:
            # Skip lane id 0

            if lane.id == 1:
                inner_lane_id = -1
                outer_lane_id = 2
            else:
                inner_lane_id = 1
                outer_lane_id = -2
            inner_neighbour_same_dir = False

        inner_neighbour_id = encode_road_section_lane_width_id(
            lane.lane_section.parentRoad.id, lane.lane_section.idx, inner_lane_id, -1
        )

        outer_neighbour_id = encode_road_section_lane_width_id(
            lane.lane_section.parentRoad.id, lane.lane_section.idx, outer_lane_id, -1
        )

        return inner_neighbour_id, outer_neighbour_id, inner_neighbour_same_dir



# -*- coding: utf-8 -*-

import abc
import numpy as np
from .eulerspiral import EulerSpiral

__author__ = "Benjamin Orthen, Stefan Urban"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


class Geometry(abc.ABC):
    """A road geometry record defines the layout of the road's reference
    line in the in the x/y-plane (plan view).

    The geometry information is split into a header which is common to all geometric elements.

    (Section 5.3.4.1 of OpenDRIVE 1.4)
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, start_position: float, heading: float, length: float):
        self._start_position = np.array(start_position)
        self._length = length
        self._heading = heading

    @property
    def start_position(self) -> float:
        """Returns the overall geometry length"""
        return self._start_position

    @property
    def length(self) -> float:
        """Returns the overall geometry length"""
        return self._length

    @property
    def heading(self) -> float:
        """Get heading of geometry.

        Returns:
          Heading, in which direction the geometry heads at start.
        """
        return self._heading

    @abc.abstractmethod
    def calc_position(self, s_pos):
        """Calculates the position of the geometry as if the starting point is (0/0)

        Args:
          s_pos:

        Returns:

        """
        return


class Line(Geometry):
    """This record describes a straight line as part of the road’s reference line.


    (Section 5.3.4.1.1 of OpenDRIVE 1.4)
    """

    def calc_position(self, s_pos):
        """

        Args:
          s_pos:

        Returns:

        """
        pos = self.start_position + np.array(
            [s_pos * np.cos(self.heading), s_pos * np.sin(self.heading)]
        )
        tangent = self.heading

        return (pos, tangent)


class Arc(Geometry):
    """This record describes an arc as part of the road’s reference line.


    (Section 5.3.4.1.3 of OpenDRIVE 1.4)
    """

    def __init__(self, start_position, heading, length, curvature):
        self.curvature = curvature
        super().__init__(start_position=start_position, heading=heading, length=length)

    def calc_position(self, s_pos):
        """

        Args:
          s_pos:

        Returns:

        """
        c = self.curvature
        hdg = self.heading - np.pi / 2

        a = 2 / c * np.sin(s_pos * c / 2)
        alpha = (np.pi - s_pos * c) / 2 - hdg

        dx = -1 * a * np.cos(alpha)
        dy = a * np.sin(alpha)

        pos = self.start_position + np.array([dx, dy])
        tangent = self.heading + s_pos * self.curvature

        return (pos, tangent)


class Spiral(Geometry):
    """This record describes a spiral as part of the road’s reference line.

    For this type of spiral, the curvature
    change between start and end of the element is linear.

    (Section 5.3.4.1.2 of OpenDRIVE 1.4)
    """

    def __init__(self, start_position, heading, length, curvStart, curvEnd):
        self._curvStart = curvStart
        self._curvEnd = curvEnd

        super().__init__(start_position=start_position, heading=heading, length=length)
        self._spiral = EulerSpiral.createFromLengthAndCurvature(
            self.length, self._curvStart, self._curvEnd
        )

    def calc_position(self, s_pos):
        """

        Args:
          s_pos:

        Returns:

        """
        (x, y, t) = self._spiral.calc(
            s_pos,
            self.start_position[0],
            self.start_position[1],
            self._curvStart,
            self.heading,
        )

        return (np.array([x, y]), t)


class Poly3(Geometry):
    """This record describes a cubic polynomial as part of the road’s reference line.


    (Section 5.3.4.1.4 of OpenDRIVE 1.4)
    """

    def __init__(self, start_position, heading, length, a, b, c, d):
        self._a = a
        self._b = b
        self._c = c
        self._d = d
        super().__init__(start_position=start_position, heading=heading, length=length)

        # raise NotImplementedError()

    def calc_position(self, s_pos):
        """

        Args:
          s_pos:

        Returns:

        """
        # TODO untested

        # Calculate new point in s_pos/t coordinate system
        coeffs = [self._a, self._b, self._c, self._d]

        t = np.polynomial.polynomial.polyval(s_pos, coeffs)

        # Rotate and translate
        srot = s_pos * np.cos(self.heading) - t * np.sin(self.heading)
        trot = s_pos * np.sin(self.heading) + t * np.cos(self.heading)

        # Derivate to get heading change
        dCoeffs = coeffs[1:] * np.array(np.arange(1, len(coeffs)))
        tangent = np.polynomial.polynomial.polyval(s_pos, dCoeffs)

        return (self.start_position + np.array([srot, trot]), self.heading + tangent)


class ParamPoly3(Geometry):
    """This record describes a parametric cubic curve as part
    of the road’s reference line in a local u/v co-ordinate system.

    This record describes an arc as part of the road’s reference line.


    (Section 5.3.4.1.5 of OpenDRIVE 1.4)
    """

    def __init__(
            self, start_position, heading, length, aU, bU, cU, dU, aV, bV, cV, dV, pRange
    ):
        super().__init__(start_position=start_position, heading=heading, length=length)

        self._aU = aU
        self._bU = bU
        self._cU = cU
        self._dU = dU
        self._aV = aV
        self._bV = bV
        self._cV = cV
        self._dV = dV

        if pRange is None:
            self._pRange = 1.0
        else:
            self._pRange = pRange

    def calc_position(self, s_pos):
        """

        Args:
          s_pos:

        Returns:

        """

        # Position
        pos = (s_pos / self.length) * self._pRange

        coeffsU = [self._aU, self._bU, self._cU, self._dU]
        coeffsV = [self._aV, self._bV, self._cV, self._dV]

        x = np.polynomial.polynomial.polyval(pos, coeffsU)
        y = np.polynomial.polynomial.polyval(pos, coeffsV)

        xrot = x * np.cos(self.heading) - y * np.sin(self.heading)
        yrot = x * np.sin(self.heading) + y * np.cos(self.heading)

        # Tangent is defined by derivation
        dCoeffsU = coeffsU[1:] * np.array(np.arange(1, len(coeffsU)))
        dCoeffsV = coeffsV[1:] * np.array(np.arange(1, len(coeffsV)))

        dx = np.polynomial.polynomial.polyval(pos, dCoeffsU)
        dy = np.polynomial.polynomial.polyval(pos, dCoeffsV)

        tangent = np.arctan2(dy, dx)

        return (self.start_position + np.array([xrot, yrot]), self.heading + tangent)



# -*- coding: utf-8 -*-

"""Provide road link classes for the OpenDRIVE implementation."""

__author__ = "Benjamin Orthen, Stefan Urban"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


class Link:
    """"""

    def __init__(self, link_id=None, predecessor=None, successor=None, neighbors=None):
        self.id_ = link_id
        self.predecessor = predecessor
        self.successor = successor
        self.neighbors = [] if neighbors is None else neighbors

    def __str__(self):
        return " > link id " + str(self._id) + " | successor: " + str(self._successor)

    @property
    def id_(self):
        """ """
        return self._id

    @id_.setter
    def id_(self, value):
        """

        Args:
          value:

        Returns:

        """
        # pylint: disable=W0201
        self._id = int(value) if value is not None else None

    @property
    def predecessor(self):
        """ """
        return self._predecessor

    @predecessor.setter
    def predecessor(self, value):
        """

        Args:
          value:

        Returns:

        """
        if not isinstance(value, Predecessor) and value is not None:
            raise TypeError("Value must be Predecessor")

        # pylint: disable=W0201
        self._predecessor = value

    @property
    def successor(self):
        """ """
        return self._successor

    @successor.setter
    def successor(self, value):
        """

        Args:
          value:

        Returns:

        """
        if not isinstance(value, Successor) and value is not None:
            raise TypeError("Value must be Successor")

        # pylint: disable=W0201
        self._successor = value

    @property
    def neighbors(self):
        """ """
        return self._neighbors

    @neighbors.setter
    def neighbors(self, value):
        """

        Args:
          value:

        Returns:

        """
        if not isinstance(value, list) or not all(
                isinstance(x, Neighbor) for x in value
        ):
            raise TypeError("Value must be list of instances of Neighbor.")

        # pylint: disable=W0201
        self._neighbors = value

    def addNeighbor(self, value):
        """

        Args:
          value:

        Returns:

        """
        if not isinstance(value, Neighbor):
            raise TypeError("Value must be Neighbor")

        self._neighbors.append(value)


class Predecessor:
    """ """

    def __init__(self, element_type=None, element_id=None, contact_point=None):
        self.elementType = element_type
        self.element_id = element_id
        self.contactPoint = contact_point

    def __str__(self):
        return (
                str(self._elementType)
                + " with id "
                + str(self._elementId)
                + " contact at "
                + str(self._contactPoint)
        )

    @property
    def elementType(self):
        """ """
        return self._elementType

    @elementType.setter
    def elementType(self, value):
        """

        Args:
          value:

        Returns:

        """
        if value not in ["road", "junction"]:
            raise AttributeError("Value must be road or junction")

        # pylint: disable=W0201
        self._elementType = value

    @property
    def element_id(self):
        """ """
        return self._elementId

    @element_id.setter
    def element_id(self, value):
        """

        Args:
          value:

        Returns:

        """
        # pylint: disable=W0201
        self._elementId = int(value)

    @property
    def contactPoint(self):
        """ """
        return self._contactPoint

    @contactPoint.setter
    def contactPoint(self, value):
        """

        Args:
          value:

        Returns:

        """
        if value not in ["start", "end"] and value is not None:
            raise AttributeError("Value must be start or end")

        # pylint: disable=W0201
        self._contactPoint = value


class Successor(Predecessor):
    """ """


class Neighbor:
    """ """

    def __init__(self, side=None, element_id=None, direction=None):
        self._side = side
        self._elementId = element_id
        self._direction = direction

    @property
    def side(self):
        """ """
        return self._side

    @side.setter
    def side(self, value):
        """

        Args:
          value:

        Returns:

        """
        if value not in ["left", "right"]:
            raise AttributeError("Value must be left or right")

        # pylint: disable=W0201
        self._side = value

    @property
    def element_id(self):
        """ """
        return self._elementId

    @element_id.setter
    def element_id(self, value):
        """

        Args:
          value:

        Returns:

        """
        # pylint: disable=W0201
        self._elementId = int(value)

    @property
    def direction(self):
        """ """
        return self._direction

    @direction.setter
    def direction(self, value):
        """

        Args:
          value:

        Returns:

        """
        if value not in ["same", "opposite"]:
            raise AttributeError("Value must be same or opposite")

        self._direction = value



# -*- coding: utf-8 -*-

import numpy as np
from lxml import etree
from .elements.opendrive import OpenDrive, Header
from .elements.road import Road
from .elements.roadLink import (
    Predecessor as RoadLinkPredecessor,
    Successor as RoadLinkSuccessor,
    Neighbor as RoadLinkNeighbor,
)
from .elements.roadtype import RoadType, Speed as RoadTypeSpeed
from .elements.roadElevationProfile import ElevationRecord as RoadElevationProfile
from .elements.roadLateralProfile import (
    Superelevation as RoadLateralProfileSuperelevation,
    Crossfall as RoadLateralProfileCrossfall,
    Shape as RoadLateralProfileShape,
)
from .elements.roadLanes import (
    LaneOffset as RoadLanesLaneOffset,
    Lane as RoadLaneSectionLane,
    LaneSection as RoadLanesSection,
    LaneWidth as RoadLaneSectionLaneWidth,
    LaneBorder as RoadLaneSectionLaneBorder,
)
from .elements.junction import (
    Junction,
    Connection as JunctionConnection,
    LaneLink as JunctionConnectionLaneLink,
)

__author__ = "Benjamin Orthen, Stefan Urban"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


def parse_opendrive(root_node) -> OpenDrive:
    """Tries to parse XML tree, returns OpenDRIVE object

    Args:
      root_node:

    Returns:
      The object representing an OpenDrive specification.

    """
    # Only accept lxml element
    if not etree.iselement(root_node):
        raise TypeError("Argument root_node is not a xml element")

    opendrive = OpenDrive()

    # Header
    header = root_node.find("header")
    if header is not None:
        parse_opendrive_header(opendrive, header)

    # Junctions
    for junction in root_node.findall("junction"):
        parse_opendrive_junction(opendrive, junction)

    # Load roads
    for road in root_node.findall("road"):
        parse_opendrive_road(opendrive, road)

    return opendrive


def parse_opendrive_road_link(newRoad, opendrive_road_link):
    """

    Args:
      newRoad:
      opendrive_road_link:

    """
    predecessor = opendrive_road_link.find("predecessor")

    if predecessor is not None:
        newRoad.link.predecessor = RoadLinkPredecessor(
            predecessor.get("elementType"),
            predecessor.get("elementId"),
            predecessor.get("contactPoint"),
        )

    successor = opendrive_road_link.find("successor")

    if successor is not None:
        newRoad.link.successor = RoadLinkSuccessor(
            successor.get("elementType"),
            successor.get("elementId"),
            successor.get("contactPoint"),
        )

    for neighbor in opendrive_road_link.findall("neighbor"):
        newNeighbor = RoadLinkNeighbor(
            neighbor.get("side"), neighbor.get("elementId"), neighbor.get("direction")
        )

        newRoad.link.neighbors.append(newNeighbor)


def parse_opendrive_road_type(road, opendrive_xml_road_type: etree.ElementTree):
    """Parse opendrive road type and append to road object.

    Args:
      road: Road to append the parsed road_type to types.
      opendrive_xml_road_type: XML element which contains the information.
      opendrive_xml_road_type: etree.ElementTree:

    """
    speed = None
    if opendrive_xml_road_type.find("speed") is not None:
        speed = RoadTypeSpeed(
            max_speed=opendrive_xml_road_type.find("speed").get("max"),
            unit=opendrive_xml_road_type.find("speed").get("unit"),
        )

    road_type = RoadType(
        s_pos=opendrive_xml_road_type.get("s"),
        use_type=opendrive_xml_road_type.get("type"),
        speed=speed,
    )
    road.types.append(road_type)


def parse_opendrive_road_geometry(newRoad, road_geometry):
    """

    Args:
      newRoad:
      road_geometry:

    """

    startCoord = [float(road_geometry.get("x")), float(road_geometry.get("y"))]

    if road_geometry.find("line") is not None:
        newRoad.planView.addLine(
            startCoord,
            float(road_geometry.get("hdg")),
            float(road_geometry.get("length")),
        )

    elif road_geometry.find("spiral") is not None:
        newRoad.planView.addSpiral(
            startCoord,
            float(road_geometry.get("hdg")),
            float(road_geometry.get("length")),
            float(road_geometry.find("spiral").get("curvStart")),
            float(road_geometry.find("spiral").get("curvEnd")),
        )

    elif road_geometry.find("arc") is not None:
        newRoad.planView.addArc(
            startCoord,
            float(road_geometry.get("hdg")),
            float(road_geometry.get("length")),
            float(road_geometry.find("arc").get("curvature")),
        )

    elif road_geometry.find("poly3") is not None:
        newRoad.planView.addPoly3(
            startCoord,
            float(road_geometry.get("hdg")),
            float(road_geometry.get("length")),
            float(road_geometry.find("poly3").get("a")),
            float(road_geometry.find("poly3").get("b")),
            float(road_geometry.find("poly3").get("c")),
            float(road_geometry.find("poly3").get("d")),
        )
        # raise NotImplementedError()

    elif road_geometry.find("paramPoly3") is not None:
        if road_geometry.find("paramPoly3").get("pRange"):

            if road_geometry.find("paramPoly3").get("pRange") == "arcLength":
                pMax = float(road_geometry.get("length"))
            else:
                pMax = None
        else:
            pMax = None

        newRoad.planView.addParamPoly3(
            startCoord,
            float(road_geometry.get("hdg")),
            float(road_geometry.get("length")),
            float(road_geometry.find("paramPoly3").get("aU")),
            float(road_geometry.find("paramPoly3").get("bU")),
            float(road_geometry.find("paramPoly3").get("cU")),
            float(road_geometry.find("paramPoly3").get("dU")),
            float(road_geometry.find("paramPoly3").get("aV")),
            float(road_geometry.find("paramPoly3").get("bV")),
            float(road_geometry.find("paramPoly3").get("cV")),
            float(road_geometry.find("paramPoly3").get("dV")),
            pMax,
        )

    else:
        raise Exception("invalid xml")


def parse_opendrive_road_elevation_profile(newRoad, road_elevation_profile):
    """

    Args:
      newRoad:
      road_elevation_profile:

    """

    for elevation in road_elevation_profile.findall("elevation"):
        newElevation = (
            RoadElevationProfile(
                float(elevation.get("a")),
                float(elevation.get("b")),
                float(elevation.get("c")),
                float(elevation.get("d")),
                start_pos=float(elevation.get("s")),
            ),
        )

        newRoad.elevationProfile.elevations.append(newElevation)


def parse_opendrive_road_lateral_profile(newRoad, road_lateral_profile):
    """

    Args:
      newRoad:
      road_lateral_profile:

    """

    for superelevation in road_lateral_profile.findall("superelevation"):
        newSuperelevation = RoadLateralProfileSuperelevation(
            float(superelevation.get("a")),
            float(superelevation.get("b")),
            float(superelevation.get("c")),
            float(superelevation.get("d")),
            start_pos=float(superelevation.get("s")),
        )

        newRoad.lateralProfile.superelevations.append(newSuperelevation)

    for crossfall in road_lateral_profile.findall("crossfall"):
        newCrossfall = RoadLateralProfileCrossfall(
            float(crossfall.get("a")),
            float(crossfall.get("b")),
            float(crossfall.get("c")),
            float(crossfall.get("d")),
            side=crossfall.get("side"),
            start_pos=float(crossfall.get("s")),
        )

        newRoad.lateralProfile.crossfalls.append(newCrossfall)

    for shape in road_lateral_profile.findall("shape"):
        newShape = RoadLateralProfileShape(
            float(shape.get("a")),
            float(shape.get("b")),
            float(shape.get("c")),
            float(shape.get("d")),
            start_pos=float(shape.get("s")),
            start_pos_t=float(shape.get("t")),
        )

        newRoad.lateralProfile.shapes.append(newShape)


def parse_opendrive_road_lane_offset(newRoad, lane_offset):
    """

    Args:
      newRoad:
      lane_offset:

    """

    newLaneOffset = RoadLanesLaneOffset(
        float(lane_offset.get("a")),
        float(lane_offset.get("b")),
        float(lane_offset.get("c")),
        float(lane_offset.get("d")),
        start_pos=float(lane_offset.get("s")),
    )

    newRoad.lanes.laneOffsets.append(newLaneOffset)


def parse_opendrive_road_lane_section(newRoad, lane_section_id, lane_section):
    """

    Args:
      newRoad:
      lane_section_id:
      lane_section:

    """

    newLaneSection = RoadLanesSection(road=newRoad)

    # Manually enumerate lane sections for referencing purposes
    newLaneSection.idx = lane_section_id

    newLaneSection.sPos = float(lane_section.get("s"))
    newLaneSection.singleSide = lane_section.get("singleSide")

    sides = dict(
        left=newLaneSection.leftLanes,
        center=newLaneSection.centerLanes,
        right=newLaneSection.rightLanes,
    )

    for sideTag, newSideLanes in sides.items():

        side = lane_section.find(sideTag)

        # It is possible one side is not present
        if side is None:
            continue

        for lane in side.findall("lane"):

            new_lane = RoadLaneSectionLane(
                parentRoad=newRoad, lane_section=newLaneSection
            )
            new_lane.id = lane.get("id")
            new_lane.type = lane.get("type")

            # In some sample files the level is not specified according to the OpenDRIVE spec
            new_lane.level = (
                "true" if lane.get("level") in [1, "1", "true"] else "false"
            )

            # Lane Links
            if lane.find("link") is not None:

                try:
                    if lane.find("link").find("predecessor") is not None:
                        new_lane.link.predecessorId = (
                            lane.find("link").find("predecessor").get("id")
                        )
                except:
                    continue

                try:
                    if lane.find("link").find("successor") is not None:
                        new_lane.link.successorId = (
                            lane.find("link").find("successor").get("id")
                        )
                except:
                    continue

            # Width
            for widthIdx, width in enumerate(lane.findall("width")):
                newWidth = RoadLaneSectionLaneWidth(
                    float(width.get("a")),
                    float(width.get("b")),
                    float(width.get("c")),
                    float(width.get("d")),
                    idx=widthIdx,
                    start_offset=float(width.get("sOffset")),
                )

                new_lane.widths.append(newWidth)

            # Border
            for borderIdx, border in enumerate(lane.findall("border")):
                newBorder = RoadLaneSectionLaneBorder(
                    float(border.get("a")),
                    float(border.get("b")),
                    float(border.get("c")),
                    float(border.get("d")),
                    idx=borderIdx,
                    start_offset=float(border.get("sOffset")),
                )

                new_lane.borders.append(newBorder)

            if lane.find("width") is None and lane.find("border") is not None:
                new_lane.widths = new_lane.borders
                new_lane.has_border_record = True

            # Road Marks
            # TODO implementation 添加roadMark
            for markIdx, roadMark in enumerate(lane.findall("roadMark")):
                newMark = {
                    "idx": markIdx,
                    "color": roadMark.get("color"),
                    "start_offset": float(roadMark.get("sOffset")),
                    # "end_offset": None,
                    "type": roadMark.get("type"),
                }
                if new_lane.road_marks:
                    new_lane.road_marks[-1]["end_offset"] = float(roadMark.get("sOffset"))
                new_lane.road_marks.append(newMark)

            # Material
            # TODO implementation

            # Visiblility
            # TODO implementation

            # Speed
            # TODO implementation

            # Access
            # TODO implementation

            # Lane Height
            # TODO implementation

            # Rules
            # TODO implementation

            newSideLanes.append(new_lane)

    newRoad.lanes.lane_sections.append(newLaneSection)


def parse_opendrive_road(opendrive, road):
    """

    Args:
      opendrive:
      road:

    """

    newRoad = Road()

    newRoad.id = int(road.get("id"))
    newRoad.name = road.get("name")

    junctionId = int(road.get("junction")) if road.get("junction") != "-1" else None

    if junctionId:
        newRoad.junction = opendrive.getJunction(junctionId)

    # TODO verify road length
    newRoad.length = float(road.get("length"))

    # Links
    opendrive_road_link = road.find("link")
    if opendrive_road_link is not None:
        parse_opendrive_road_link(newRoad, opendrive_road_link)

    # Type
    for opendrive_xml_road_type in road.findall("type"):
        parse_opendrive_road_type(newRoad, opendrive_xml_road_type)

    # Plan view
    for road_geometry in road.find("planView").findall("geometry"):
        parse_opendrive_road_geometry(newRoad, road_geometry)

    # Elevation profile
    road_elevation_profile = road.find("elevationProfile")
    if road_elevation_profile is not None:
        parse_opendrive_road_elevation_profile(newRoad, road_elevation_profile)

    # Lateral profile
    road_lateral_profile = road.find("lateralProfile")
    if road_lateral_profile is not None:
        parse_opendrive_road_lateral_profile(newRoad, road_lateral_profile)

    # Lanes
    lanes = road.find("lanes")

    if lanes is None:
        raise Exception("Road must have lanes element")

    # Lane offset
    for lane_offset in lanes.findall("laneOffset"):
        parse_opendrive_road_lane_offset(newRoad, lane_offset)

    # Lane sections
    for lane_section_id, lane_section in enumerate(
            road.find("lanes").findall("laneSection")
    ):
        parse_opendrive_road_lane_section(newRoad, lane_section_id, lane_section)

    # Objects
    # TODO implementation

    # Signals
    # TODO implementation
    calculate_lane_section_lengths(newRoad)

    opendrive.roads.append(newRoad)


def calculate_lane_section_lengths(newRoad):
    """

    Args:
      newRoad:

    """
    # OpenDRIVE does not provide lane section lengths by itself, calculate them by ourselves
    for lane_section in newRoad.lanes.lane_sections:

        # Last lane section in road
        if lane_section.idx + 1 >= len(newRoad.lanes.lane_sections):
            lane_section.length = newRoad.planView.length - lane_section.sPos

        # All but the last lane section end at the succeeding one
        else:
            lane_section.length = (
                    newRoad.lanes.lane_sections[lane_section.idx + 1].sPos
                    - lane_section.sPos
            )

    # OpenDRIVE does not provide lane width lengths by itself, calculate them by ourselves
    for lane_section in newRoad.lanes.lane_sections:
        for lane in lane_section.allLanes:
            widthsPoses = np.array(
                [x.start_offset for x in lane.widths] + [lane_section.length]
            )
            widthsLengths = widthsPoses[1:] - widthsPoses[:-1]

            for widthIdx, width in enumerate(lane.widths):
                width.length = widthsLengths[widthIdx]


def parse_opendrive_header(opendrive, header):
    """

    Args:
      opendrive:
      header:

    """

    parsed_header = Header(
        header.get("revMajor"),
        header.get("revMinor"),
        header.get("name"),
        header.get("version"),
        header.get("date"),
        header.get("north"),
        header.get("south"),
        header.get("west"),
        header.get("vendor"),
    )
    # Reference
    if header.find("geoReference") is not None:
        parsed_header.geo_reference = header.find("geoReference").text

    opendrive.header = parsed_header


def parse_opendrive_junction(opendrive, junction):
    """

    Args:
      opendrive:
      junction:

    """
    newJunction = Junction()

    newJunction.id = int(junction.get("id"))
    newJunction.name = str(junction.get("name"))

    for connection in junction.findall("connection"):

        newConnection = JunctionConnection()

        newConnection.id = connection.get("id")
        newConnection.incomingRoad = connection.get("incomingRoad")
        newConnection.connectingRoad = connection.get("connectingRoad")
        if connection.get("contactPoint") not in ["start", "end"]:
            continue
        newConnection.contactPoint = connection.get("contactPoint")

        for laneLink in connection.findall("laneLink"):
            newLaneLink = JunctionConnectionLaneLink()

            newLaneLink.fromId = laneLink.get("from")
            newLaneLink.toId = laneLink.get("to")

            newConnection.addLaneLink(newLaneLink)

        newJunction.addConnection(newConnection)

    opendrive.junctions.append(newJunction)



# -*- coding: utf-8 -*-

__author__ = "Benjamin Orthen, Stefan Urban"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


class Junction:
    """ """

    # TODO priority
    # TODO controller

    def __init__(self):
        self._id = None
        self._name = None
        self._connections = []

    @property
    def id(self):
        """ """
        return self._id

    @id.setter
    def id(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._id = int(value)

    @property
    def name(self):
        """ """
        return self._name

    @name.setter
    def name(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._name = str(value)

    @property
    def connections(self):
        """ """
        return self._connections

    def addConnection(self, connection):
        """

        Args:
          connection:

        Returns:

        """
        if not isinstance(connection, Connection):
            raise TypeError("Has to be of instance Connection")

        self._connections.append(connection)


class Connection:
    """ """

    def __init__(self):
        self._id = None
        self._incomingRoad = None
        self._connectingRoad = None
        self._contactPoint = None
        self._laneLinks = []

    @property
    def id(self):
        """ """
        return self._id

    @id.setter
    def id(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._id = int(value)

    @property
    def incomingRoad(self):
        """ """
        return self._incomingRoad

    @incomingRoad.setter
    def incomingRoad(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._incomingRoad = int(value)

    @property
    def connectingRoad(self):
        """ """
        return self._connectingRoad

    @connectingRoad.setter
    def connectingRoad(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._connectingRoad = int(value)

    @property
    def contactPoint(self):
        """ """
        return self._contactPoint

    @contactPoint.setter
    def contactPoint(self, value):
        """

        Args:
          value:

        Returns:

        """
        if value not in ["start", "end"]:
            raise AttributeError("Contact point can only be start or end.")

        self._contactPoint = value

    @property
    def laneLinks(self):
        """ """
        return self._laneLinks

    def addLaneLink(self, laneLink):
        """

        Args:
          laneLink:

        Returns:

        """
        if not isinstance(laneLink, LaneLink):
            raise TypeError("Has to be of instance LaneLink")

        self._laneLinks.append(laneLink)


class LaneLink:
    """ """

    def __init__(self):
        self._from = None
        self._to = None

    def __str__(self):
        return str(self._from) + " > " + str(self._to)

    @property
    def fromId(self):
        """ """
        return self._from

    @fromId.setter
    def fromId(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._from = int(value)

    @property
    def toId(self):
        """ """
        return self._to

    @toId.setter
    def toId(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._to = int(value)



import math
import collections
from .config import border_line_width, convert_attribute_mapping
from .functions import qtpoint2point, xyz2xzy, create_curve, chunk, calc_boundary


# 获取 路网三角形
def get_triangle_polygon(netiface):
    polygon_info = collections.defaultdict(list)

    # 绘制路面(黑色)
    for link in netiface.links():
        left_points = qtpoint2point(link.leftBreakPoint3Ds())
        right_points = qtpoint2point(link.rightBreakPoint3Ds())
        polygon_info["black"] += create_curve(left_points, right_points)
    for connector in netiface.connectors():
        for laneConnector in connector.laneConnectors():
            left_points = qtpoint2point(laneConnector.leftBreakPoint3Ds())
            # 防止 nan 情况发生(长度为0的情况)
            left_points = [_ for _ in left_points if not math.isnan(_[2])]
            right_points = qtpoint2point(laneConnector.rightBreakPoint3Ds())
            polygon_info["black"] += create_curve(left_points, right_points)

    # 绘制左右边界线(黄色实线，白色实线，白色虚线)
    for link in netiface.links():
        lanes = list(link.lanes())
        for index, lane in enumerate(lanes):
            if index == 0:
                # 最右侧车道绘制右侧边界(白色实线)
                base_points = qtpoint2point(link.rightBreakPoint3Ds())
                left_points, right_points = calc_boundary(base_points, border_line_width)
                polygon_info['white'] += create_curve(left_points, right_points)
            # 所有车道绘制左侧边界
            if lane == lanes[-1]:
                # 最左侧车道绘制黄色实线
                base_points = qtpoint2point(link.leftBreakPoint3Ds())
                left_points, right_points = calc_boundary(base_points, border_line_width)
                polygon_info['yellow'] += create_curve(left_points, right_points)
            elif lane.actionType() != lanes[index+1].actionType():
                # 左侧相邻车道类型不一致，绘制白色实线
                base_points = qtpoint2point(link.leftBreakPoint3Ds())
                left_points, right_points = calc_boundary(base_points, border_line_width)
                polygon_info['white'] += create_curve(left_points, right_points)
            else:
                # # 左侧相邻车道类型一致，绘制白色虚线
                base_points = qtpoint2point(link.leftBreakPoint3Ds())
                left_points, right_points = calc_boundary(base_points, border_line_width)
                polygon_info['white'] += create_curve(left_points, right_points, split=True)
    
    return polygon_info


def get_info(netiface):
    # 获取三角形数据
    polygon_data = get_triangle_polygon(netiface)

    # 存储最终的unity三角形信息
    unity_triangle_data = collections.defaultdict(list)

    for color, triangles in polygon_data.items():
        unity_attribute = convert_attribute_mapping[color]
        # 存储最终的 unity 数据
        for triangle in triangles:
            # 遍历每一个原始三角形点位
            for point in triangle:
                # unity 坐标系问题，需要 xyz 转 xzy, 同时将三维列表转换为一维列表
                new_point = xyz2xzy(point)
                # unity 是一个二维数组，元素为每个点的坐标，默认每三个点构成一个三角形
                unity_triangle_data[unity_attribute].append(new_point)

    # 转换为 unity数据
    unity_count = {}
    for key, value in unity_triangle_data.items():
        # unity 单次接收的三角形数量有限制，不能超过 256 * 256 个
        unity_triangle_data[key] = [
            {'pointsArray': info, 'drawOrder': [i for i in range(len(info))], 'count': int(len(info))}
            for info in chunk(value, 60000)]
        unity_count[key] = len(unity_triangle_data[key])

    # 三角形数量
    unity_data = {'unity': unity_triangle_data, 'count': {}}
    for k, v in unity_data['unity'].items():
        unity_data['count'][k] = len(v)

    return unity_data


import os
import time
import json
from datetime import datetime
from queue import Queue
from threading import Thread
from .kafka_producer import KafkaMessageProducer


class SendData:
    def __init__(self, traj_json_config:str=None, traj_kafka_config:dict=None):
        self.json_save_path = None
        self.kafka_producer = None

        self.set_config(traj_json_config, traj_kafka_config)

        self.send_data_queue = Queue()
        self.send_data_flag = True
        self.send_data_process = Thread(target=self.send_data)
        self.send_data_process.start()

    def set_config(self, traj_json_config:str, traj_kafka_config:dict):
        # Json
        if traj_json_config:
            # 获取文件夹名称
            current_time = datetime.now().strftime("%Y%m%d%H%M%S")
            folder_path = os.path.join(traj_json_config, f"轨迹数据_{current_time}")
            # 创建文件夹
            os.makedirs(folder_path, exist_ok=True)
            self.json_save_path = os.path.join(folder_path, "{}.json")

        # Kakfa
        if traj_kafka_config:
            ip = traj_kafka_config["ip"]
            port = traj_kafka_config["port"]
            topic = traj_kafka_config["topic"]
            self.kafka_producer = KafkaMessageProducer(f"{ip}:{port}", topic)

    def put_data(self, traj_data:dict):
        self.send_data_queue.put(traj_data)

    def send_data(self):
        print("数据发送线程已启动")

        while True:
            time.sleep(0.01)
            if self.send_data_queue.empty():
                if self.send_data_flag:
                    continue
                else:
                    print("数据发送线程已关闭")
                    break

            try:
                traj_data = self.send_data_queue.get_nowait()
            except:
                continue

            # 当前仿真计算批次
            batchNum = traj_data["batchNum"]

            t1 = time.time()
            # 需要保存为json
            if self.json_save_path:
                # 当前仿真计算批次
                file_path = self.json_save_path.format(batchNum)
                # 将JSON数据写入文件
                with open(file_path, 'w', encoding="utf-8") as file:
                    json.dump(traj_data, file, indent=4, ensure_ascii=False)

            t2 = time.time()
            # 需要上传至kafka
            if self.kafka_producer:
                traj_data_json = json.dumps(traj_data)
                self.send_data_flag = self.kafka_producer.send_message(traj_data_json)
                if not self.send_data_flag:
                    print("由于Kafka数据发送失败，数据发送线程关闭")
                    break

            t3 = time.time()
            json_time = round((t2 - t1) * 1000, 1)
            kafka_time = round((t3 - t2) * 1000, 1)
            print(f"仿真批次：{batchNum}，导出时间：{json_time}ms，上传时间：{kafka_time}ms，队列大小：{self.send_data_queue.qsize()}")

    def close(self):
        # 发送数据线程关闭
        self.send_data_flag = False

        # 清空队列
        while not self.send_data_queue.empty():
            time.sleep(0.01)
        print("数据队列已清空")

        # Kafka
        # if self.kafka_producer is not None:
        #     self.kafka_producer.close()
        #     print("Kafka已关闭")



